# --------------------------------------------------------------------------#
# Copyright (C) 2015 - Present by Tibit Communications, Inc.                #
# All rights reserved.                                                      #
#                                                                           #
#    _______ ____  _ ______                                                 #
#   /_  __(_) __ )(_)_  __/                                                 #
#    / / / / __  / / / /                                                    #
#   / / / / /_/ / / / /                                                     #
#  /_/ /_/_____/_/ /_/                                                      #
#                                                                           #
# --------------------------------------------------------------------------#

from voltha_protos.events_pb2 import EventCategory, EventSubCategory
from voltha_protos.common_pb2 import OperStatus

from pyvoltha_min.adapters.extensions.events.adapter_events import DeviceEventBase


class OltPonInterfaceDownEvent(DeviceEventBase):
    event_name = 'OLT_PON_INTERFACE_DOWN'

    def __init__(self, event_mgr, intf_id, oper_status, raised_ts):
        super().__init__(event_mgr, raised_ts, object_type='OLT Pon Interface Down',
                         event=self.event_name,
                         category=EventCategory.COMMUNICATION,
                         sub_category=EventSubCategory.OLT)
        # Added port type to indicate if alarm was on NNI or PON
        self._intf_id = str(intf_id)
        self._oper_status = 'up' if oper_status == OperStatus.ACTIVE else 'down'

    def get_context_data(self):
        """ Strangly enough, the PON ID is not encoded into this message in the OpenOLT """
        return {'intf-id:': str(self._intf_id),
                'oper-state': self._oper_status}

    def to_dict(self):
        """ Serialize to a dictionary """
        data = super().to_dict()
        data['int_id'] = self._intf_id
        data['oper_status'] = self._oper_status
        return data

    @staticmethod
    def from_dict(data, event_mgr):
        return OltPonInterfaceDownEvent(event_mgr, data['int_id'],
                                        data['oper_status'], data['raised_ts'])



# Copyright 2020 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import structlog
import etcd3
import etcd3.utils as utils

from twisted.internet import threads
from twisted.internet.defer import CancelledError, inlineCallbacks, returnValue
from twisted.python.failure import Failure

DEFAULT_KVSTORE_RETRIES = 0


class TwistedEtcdStore:
    def __init__(self, host, port, path_prefix, timeout=None, default_retries=DEFAULT_KVSTORE_RETRIES):
        self._etcd = etcd3.client(host=host, port=port, timeout=timeout)
        self._host = host
        self._port = port
        self._path_prefix = path_prefix
        self._default_retries = default_retries
        self._timeout = timeout

        self.log = structlog.get_logger(client='{}:{}'.format(self._host, self._port))
        # TODO: Before deprecating the blocking 'EtcdStore' class, add support
        #       for retries. Specifically look for a failed connection where a
        #       a new client needs to be added.

    def __str__(self):
        return 'etcd-{}:{}/{}, Timeout: {}'.format(self._host, self._port,
                                                   self._path_prefix, self._timeout)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @inlineCallbacks
    def __getitem__(self, key):
        (value, _meta) = yield self.get(self.make_path(key))
        if value is not None:
            returnValue(value)

        raise KeyError(key)

    @inlineCallbacks
    def __contains__(self, key):
        (value, _meta) = yield self.get(key)
        returnValue(value is not None)

    @inlineCallbacks
    def __setitem__(self, key, value):
        _results = yield self.set(key, value)

    @inlineCallbacks
    def __delitem__(self, key):
        _results = yield self.delete(key)

    def close(self):
        client, self._etcd = self._etcd, None
        if client is not None:
            threads.deferToThread(client.close)

    def make_path(self, key):
        if key is None:
            return self._path_prefix
        return '{}/{}'.format(self._path_prefix, key)

    def _failure(self, reason, key, operation, retries):
        """ Common error handler """
        if isinstance(reason, Failure) and issubclass(type(reason.value), CancelledError):
            self.log.debug('{}-cancelled'.format(operation))
        else:
            self.log.info('{}-failure'.format(operation), error=reason, key=key)

            if retries:
                # TODO: Need to retry and set up *args, **kwargs appropriately
                #       for the command.  Watch specifically for client connection
                #       failures and test best way to handle this.   See also if
                #       timeouts are supported...  The older EtcdStore() that blocked
                #       had a small pause on the retry mechanism.
                retries -= 1
                self.log.debug('retries-not-yet-supported', remaining=retries)
                # TODO: what is best way to handle callback/errback chain and can it even be
                #       done?  May just need to watch for connection failures and determine
                #       the best way to recover.

        return reason

    def get(self, key, retries=None):
        def success(results):
            (value, _meta) = results
            return value

        deferred = threads.deferToThread(self._etcd.get, self.make_path(key))
        deferred.addCallbacks(success, self._failure,
                              errbackArgs=[key, 'get', retries or self._default_retries])
        return deferred

    def set(self, key, value, retries=None):

        def success(results):
            if results:
                return results
            return False

        deferred = threads.deferToThread(self._etcd.put, self.make_path(key), value)
        deferred.addCallbacks(success, self._failure,
                              errbackArgs=[key, 'set', retries or self._default_retries])
        return deferred

    def list(self, key, keys_only=False, retries=None):

        def success(results):
            if results:
                return results
            return False

        deferred = threads.deferToThread(self._etcd.get_prefix, self.make_path(key),
                                         keys_only=keys_only)
        deferred.addCallbacks(success, self._failure,
                              errbackArgs=[key, 'list', retries or self._default_retries])
        return deferred

    def watch(self, key, callback, watch_prefix=False, retries=None):

        def success(results):
            return results

        path = self.make_path(key)
        if watch_prefix:
            kwargs = dict(range_end=utils.increment_last_byte(utils.to_bytes(path)))
            deferred = threads.deferToThread(self._etcd.add_watch_callback, path, callback, **kwargs)
        else:
            deferred = threads.deferToThread(self._etcd.add_watch_callback, path, callback)

        deferred.addCallbacks(success, self._failure,
                              errbackArgs=[key, 'watch', retries or self._default_retries])
        return deferred

    def delete(self, key, retries=None):
        self.log.debug('entry', key=key)

        if key is None or len(key) == 0:
            raise ValueError("key-not-provided: '{}'".format(key))

        def success(results, k):
            self.log.debug('delete-success', results=results, key=k)
            if results:
                return results
            return False

        deferred = threads.deferToThread(self._etcd.delete, self.make_path(key))
        deferred.addCallbacks(success, self._failure, callbackArgs=[key],
                              errbackArgs=[key, 'delete', retries or self._default_retries])
        return deferred

    def delete_prefix(self, prefix, retries=None):
        self.log.debug('entry', prefix=prefix)

        def success(results, k):
            self.log.debug('delete-prefix-success', results=results, prefix=k)
            if results:
                return results
            return False

        if prefix is not None and len(prefix) > 0:
            prefix = self.make_path(prefix)
        else:
            # Deleting a range of keys with a prefix
            prefix = self._path_prefix
            while prefix[-1:] == '/':
                prefix = prefix[:-1]

        deferred = threads.deferToThread(self._etcd.delete_prefix, prefix)
        deferred.addCallbacks(success, self._failure, callbackArgs=[prefix],
                              errbackArgs=[prefix, 'delete-prefix', retries or self._default_retries])
        return deferred



#
# Copyright 2020 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import structlog
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue

from .common.kvstore.twisted_etcd_store import TwistedEtcdStore         # pylint: disable=relative-beyond-top-level
from ..common.structlog_setup import string_to_int                      # pylint: disable=relative-beyond-top-level

GLOBAL_CONFIG_ROOT_NODE = "global"
DEFAULT_KV_STORE_CONFIG_PATH = "config"
KV_STORE_DATA_PATH_PREFIX = "service/voltha"
KV_STORE_PATH_SEPARATOR = "/"
CONFIG_TYPE = "loglevel"
DEFAULT_PACKAGE_NAME = "default"
DEFAULT_SUFFIX = KV_STORE_PATH_SEPARATOR + DEFAULT_PACKAGE_NAME
GLOBAL_DEFAULT_LOGLEVEL = b"WARN"
LOG_LEVEL_INT_NOT_FOUND = 0


class LogController:
    instance_id = None
    active_log_level = None

    def __init__(self, etcd_host, etcd_port, component_name, watch_callback=None):
        self.log = structlog.get_logger()
        self.etcd_host = etcd_host
        self.etcd_port = etcd_port
        self.etcd_client = TwistedEtcdStore(self.etcd_host, self.etcd_port, KV_STORE_DATA_PATH_PREFIX)
        self.global_config_path = self.make_component_path(GLOBAL_CONFIG_ROOT_NODE) + DEFAULT_SUFFIX
        self.component_config_path = self.make_component_path(component_name)
        self.global_log_level = GLOBAL_DEFAULT_LOGLEVEL
        self._watch_callback = watch_callback
        self._component_name = component_name

    @staticmethod
    def make_component_path(component):
        return DEFAULT_KV_STORE_CONFIG_PATH + KV_STORE_PATH_SEPARATOR + component + \
               KV_STORE_PATH_SEPARATOR + CONFIG_TYPE

    def make_config_path(self, key):
        return self.component_config_path + KV_STORE_PATH_SEPARATOR + key

    def is_global_config_path(self, key):
        if isinstance(key, bytes):
            key = key.decode('utf-8')

        return key is not None and \
               key[:len(KV_STORE_DATA_PATH_PREFIX)] == KV_STORE_DATA_PATH_PREFIX and \
               key[len(KV_STORE_DATA_PATH_PREFIX)+1:] == self.global_config_path

    def is_component_default_path(self, key):
        if isinstance(key, bytes):
            key = key.decode('utf-8')

        default_path = self.make_config_path(DEFAULT_PACKAGE_NAME)
        return key is not None and \
               key[:len(KV_STORE_DATA_PATH_PREFIX)] == KV_STORE_DATA_PATH_PREFIX and \
               key[len(KV_STORE_DATA_PATH_PREFIX)+1:] == default_path

    @inlineCallbacks
    def get_global_loglevel(self):
        loglevel = self.global_log_level
        try:
            level = yield self.etcd_client.get(self.global_config_path)
            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))

                if level_int == 0:
                    self.log.warn("unsupported-log-level", requested_level=level)
                else:
                    loglevel = level

        except KeyError:
            self.log.warn("failed-to-retrieve-log-level", path=self.global_config_path)

        returnValue(loglevel)

    @inlineCallbacks
    def get_component_default_loglevel(self, global_default_loglevel):
        self.log.info("entry", path=self.component_config_path)
        component_default_loglevel = global_default_loglevel

        try:
            path = self.make_config_path(DEFAULT_PACKAGE_NAME)
            level = yield self.etcd_client.get(path)

            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))

                if level_int == 0:
                    self.log.warn("unsupported-log-level", unsupported_level=level)

                else:
                    component_default_loglevel = level
                    self.log.info("default-loglevel", new_level=level)

        except KeyError:
            self.log.warn("failed-to-retrieve-log-level",
                          path=self.component_config_path + DEFAULT_SUFFIX)

        if component_default_loglevel == "":
            component_default_loglevel = GLOBAL_DEFAULT_LOGLEVEL.encode('utf-8')

        returnValue(component_default_loglevel)

    @inlineCallbacks
    def get_package_loglevel_int(self, package):
        self.log.debug("entry", path=self.component_config_path)
        level_int = LOG_LEVEL_INT_NOT_FOUND
        path = self.make_config_path(package)
        try:
            level = yield self.etcd_client.get(path)
            if level is not None:
                level_int = string_to_int(str(level, 'utf-8'))
                if level_int == 0:
                    self.log.warn("unsupported-log-level", unsupported_level=level)
                    level_int = LOG_LEVEL_INT_NOT_FOUND

        except KeyError:
            self.log.warn("failed-to-retrieve-log-level", path=path)

        returnValue(level_int)

    @inlineCallbacks
    def start_watch_log_config_change(self, instance_id, initial_default_loglevel):
        self.log.debug("entry")
        LogController.instance_id = instance_id

        if self._component_name is None:
            raise Exception("Invalid pod component name")

        _results = yield self.etcd_client.watch(self.global_config_path, self.watch_callback)
        _results = yield self.etcd_client.watch(self.component_config_path, self.watch_callback,
                                                watch_prefix=True)

        _results = yield self.set_default_loglevel(self.global_config_path,
                                                   self.component_config_path,
                                                   initial_default_loglevel.upper())

    def watch_callback(self, event):
        reactor.callFromThread(self.process_log_config_change, event)

    @inlineCallbacks
    def process_log_config_change(self, watch_event=None):
        # Forward to user callback
        if self._watch_callback is not None:
            results = yield self._watch_callback(watch_event)
            returnValue(results)

        returnValue(False)

    @inlineCallbacks
    def set_default_loglevel(self, global_config_path, component_config_path, initial_default_loglevel):
        self.log.debug('set-default-level', global_path=global_config_path,
                       component_path=component_config_path,
                       def_level=initial_default_loglevel)

        if (yield self.etcd_client.get(global_config_path)) is None:
            _results = yield self.etcd_client.set(global_config_path, GLOBAL_DEFAULT_LOGLEVEL)

        default_path = self.make_config_path(DEFAULT_PACKAGE_NAME)
        if (yield self.etcd_client.get(default_path)) is None:
            _results = yield self.etcd_client.set(default_path, initial_default_loglevel)



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: skip-file

import threading

from confluent_kafka import Consumer
from confluent_kafka import Producer as _kafkaProducer
from structlog import get_logger
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue, gatherResults, succeed
from twisted.internet.threads import deferToThread
from zope.interface import implementer

from pyvoltha_min.common.utils.registry import IComponent
from .event_bus_publisher import EventBusPublisher

log = get_logger()


@implementer(IComponent)
class KafkaProxy(object):
    """
    This is a singleton proxy kafka class to hide the kafka client details. This
    proxy uses confluent-kafka-python as the kafka client. Since that client is
    not a Twisted client then requests to that client are wrapped with
    twisted.internet.threads.deferToThread to avoid any potential blocking of
    the Twisted loop.
    """
    _kafka_instance = None

    def __init__(self,
                 kafka_endpoint='localhost:9092',
                 ack_timeout=1000,
                 max_req_attempts=10,
                 consumer_poll_timeout=10,
                 config=None):
        config = config or {}
        # return an exception if the object already exist
        if KafkaProxy._kafka_instance:
            raise Exception('Singleton exist for :{}'.format(KafkaProxy))

        log.debug('initializing', endpoint=kafka_endpoint)
        self.ack_timeout = ack_timeout
        self.max_req_attempts = max_req_attempts
        self.kafka_endpoint = kafka_endpoint
        self.config = config
        self.kclient = None
        self.kproducer = None
        self.kproducer_heartbeat = None
        self.alive_state_handler = None
        self.event_bus_publisher = None
        self.stopping = False
        self.faulty = False
        self.alive = False
        self.consumer_poll_timeout = consumer_poll_timeout
        self.topic_consumer_map = {}
        self.topic_callbacks_map = {}
        self.topic_any_map_lock = threading.Lock()
        log.debug('initialized', endpoint=kafka_endpoint)

    @inlineCallbacks
    def start(self):
        log.debug('starting')
        self._get_kafka_producer()
        KafkaProxy._kafka_instance = self
        self.event_bus_publisher = yield EventBusPublisher(
            self, self.config.get('event_bus_publisher', {})).start()
        log.info('started')
        KafkaProxy.faulty = False
        self.stopping = False
        self.alive = True
        returnValue(self)

    def stop(self):
        d = succeed(True)
        try:
            log.debug('stopping-kafka-proxy')
            self.stopping = True
            self.alive = False
            dl = []
            client, self.kclient = self.kclient, None
            producer, self.kproducer = self.kproducer, None
            if client:
                try:
                    log.debug('stopping-kclient-kafka-proxy')
                    d = deferToThread(client.close)
                    d.addTimeout(0.3, reactor, lambda _: log.error('client-timeout'))
                    d.addCallbacks(lambda _: log.debug('client-success'),
                                   lambda _: log.error('client-failure'))
                    dl.append(d)

                except Exception as e:
                    log.exception('failed-stopped-kclient-kafka-proxy', e=e)

            if producer:
                try:
                    log.debug('stopping-kproducer-kafka-proxy')
                    d = deferToThread(producer.flush)
                    d.addTimeout(0.3, reactor, lambda _: log.error('producer-timeout'))
                    d.addCallbacks(lambda _: log.debug('producer-success'),
                                   lambda _: log.error('producer-failure'))
                    dl.append(d)
                except Exception as e:
                    log.exception('failed-stopped-kproducer-kafka-proxy', e=e)

            # Stop all consumers
            try:
                self.topic_any_map_lock.acquire()
                log.debug('stopping-consumers-kafka-proxy', size=len(self.topic_consumer_map))

                consumer_map, self.topic_consumer_map = self.topic_consumer_map, dict()
                for _, c in consumer_map.items():
                    d = deferToThread(c.close)
                    d.addTimeout(0.3, reactor, lambda _: log.error('consumer-timeout'))
                    d.addCallbacks(lambda _: log.debug('consumer-success'),
                                   lambda _: None)
                    # lambda reason: log.error('consumer-failure: {}'.format(str(reason))))
                    dl.append(d)

                self.topic_callbacks_map.clear()
                log.debug('stopped-consumers-kafka-proxy')

            except Exception as e:
                log.exception('failed-stopped-consumers-kafka-proxy', e=e)

            finally:
                self.topic_any_map_lock.release()
                log.debug('stopping-consumers-kafka-proxy-released-lock')

            if len(dl) > 0:
                try:
                    log.info('client-producer-wait', size=len(dl))
                    d = gatherResults(dl, consumeErrors=True)
                    d.addTimeout(0.5, reactor, lambda _: None)
                    d.addCallbacks(lambda _: log.debug('client-wait-success'),
                                   lambda _: log.error('client-wait-failure'))

                except Exception as e:
                    log.exception('client-producer-wait-failed', e=e)

            # TODO: Why is this commented out?  Is it still needed in v2.x?
            # try:
            #    if self.event_bus_publisher:
            #        yield self.event_bus_publisher.stop()
            #        self.event_bus_publisher = None
            #        log.debug('stopped-event-bus-publisher-kafka-proxy')
            # except Exception, e:
            #    log.debug('failed-stopped-event-bus-publisher-kafka-proxy')
            #    pass

            log.debug('stopped-kafka-proxy')

        except Exception as e:
            self.kclient = None
            self.kproducer = None
            # success = False
            # self.event_bus_publisher = None
            log.exception('failed-stopped-kafka-proxy', e=e)

        return d

    def _get_kafka_producer(self):

        try:
            _k_endpoint = self.kafka_endpoint
            self.kproducer = _kafkaProducer({'bootstrap.servers': _k_endpoint, })

        except Exception as e:
            log.exception('failed-get-kafka-producer', e=e)
            return

    @inlineCallbacks
    def _wait_for_messages(self, consumer, topic):
        log.debug('entry', topic=topic)
        while True:
            try:
                msg = yield deferToThread(consumer.poll,
                                          self.consumer_poll_timeout)

                if self.stopping:
                    log.debug("stop-request-received", topic=topic)
                    break

                log.debug('rx', topic=topic)
                if msg is None:
                    continue
                if msg.error():
                    # This typically is received when there are no more messages
                    # to read from kafka. Ignore.
                    continue

                # Invoke callbacks
                log.debug('invoking callbacks', topic=topic, count=len(self.topic_callbacks_map))
                for cb in self.topic_callbacks_map[topic]:
                    yield cb(msg)
            except Exception as e:
                log.debug("exception-receiving-msg", topic=topic, e=e)

    @inlineCallbacks
    def subscribe(self, topic, callback, groupId, offset='latest'):
        """
        subscribe allows a caller to subscribe to a given kafka topic.  This API
        always create a group consumer.
        :param topic - the topic to subscribe to
        :param callback - the callback to invoke whenever a message is received
        on that topic
        :param groupId - the groupId for this consumer.  In the current
        implementation there is a one-to-one mapping between a topic and a
        groupId.  In other words, once a groupId is used for a given topic then
        we won't be able to create another groupId for the same topic.
        :param offset:  the kafka offset from where the consumer will start
        consuming messages
        """
        log.debug('entry', topic=topic, groupId=groupId)
        try:
            self.topic_any_map_lock.acquire()
            if topic in self.topic_consumer_map:
                # Just add the callback
                if topic in self.topic_callbacks_map:
                    self.topic_callbacks_map[topic].append(callback)
                else:
                    self.topic_callbacks_map[topic] = [callback]
                return

            # Create consumer for that topic
            c = Consumer({
                'bootstrap.servers': self.kafka_endpoint,
                'group.id': groupId,
                'auto.offset.reset': offset
            })
            log.debug('sending-to-thread')
            yield deferToThread(c.subscribe, [topic])
            # c.subscribe([topic])
            self.topic_consumer_map[topic] = c
            self.topic_callbacks_map[topic] = [callback]
            # Start the consumer
            reactor.callLater(0, self._wait_for_messages, c, topic)
        except Exception as e:
            log.exception("topic-subscription-error", e=e)
        finally:
            self.topic_any_map_lock.release()

    @inlineCallbacks
    def unsubscribe(self, topic, callback):
        """
        Unsubscribe to a given topic.  Since there they be multiple callers
        consuming from the same topic then to ensure only the relevant caller
        gets unsubscribe then the callback is used as a differentiator.   The
        kafka consumer will be closed when there are no callbacks required.
        :param topic: topic to unsubscribe
        :param callback: callback the caller used when subscribing to the topic.
        If multiple callers have subscribed to a topic using the same callback
        then the first callback on the list will be removed.
        :return:None
        """
        try:
            self.topic_any_map_lock.acquire()
            log.debug("unsubscribing-to-topic", topic=topic)
            if topic in self.topic_callbacks_map:
                index = 0
                for cb in self.topic_callbacks_map[topic]:
                    if cb == callback:
                        break
                    index += 1
                if index < len(self.topic_callbacks_map[topic]):
                    self.topic_callbacks_map[topic].pop(index)

                if len(self.topic_callbacks_map[topic]) == 0:
                    # Stop the consumer
                    if topic in self.topic_consumer_map:
                        yield deferToThread(
                            self.topic_consumer_map[topic].close)
                        del self.topic_consumer_map[topic]
                    del self.topic_callbacks_map[topic]
                    log.debug("unsubscribed-to-topic", topic=topic)
                else:
                    log.debug("consumers-for-topic-still-exist", topic=topic,
                              num=len(self.topic_callbacks_map[topic]))
        except Exception as e:
            log.exception("topic-unsubscription-error", e=e)
        finally:
            self.topic_any_map_lock.release()
            log.debug("unsubscribing-to-topic-release-lock", topic=topic)

    @inlineCallbacks
    def send_message(self, topic, msg, key=None, span=None):
        # first check whether we have a kafka producer.
        error = None
        try:
            if self.faulty is False:
                if self.kproducer is None:
                    self._get_kafka_producer()
                    # Lets the next message request do the retry if still a failure
                    if self.kproducer is None:
                        log.error('no-kafka-producer',
                                  endpoint=self.kafka_endpoint)
                        return

                log.debug('sending-kafka-msg', topic=topic)
                # msgs = [msg]

                if self.kproducer is not None and self.event_bus_publisher and self.faulty is False:
                    d = deferToThread(self.kproducer.produce, topic, msg, key)
                    yield d
                    log.debug('sent-kafka-msg', topic=topic)
                    # send a lightweight poll to avoid an exception after 100k messages.
                    d1 = deferToThread(self.kproducer.poll, 0)
                    yield d1
                else:
                    return

        except Exception as e:
            self.faulty = True
            self.alive_state_handler.callback(self.alive)
            error = 'failed to send kafka-msg'
            log.error(error, topic=topic, e=e)

            # set the kafka producer to None.  This is needed if the
            # kafka docker went down and comes back up with a different
            # port number.
            if self.stopping is False:
                log.debug('stopping-kafka-proxy')
                self.alive = False
                try:
                    self.stopping = True
                    self.stop()
                    self.stopping = False
                    log.debug('stopped-kafka-proxy')

                except Exception as e:
                    log.exception('failed-stopping-kafka-proxy', e=e)
            else:
                log.info('already-stopping-kafka-proxy')
            return

        finally:
            if span is not None:
                if error:
                    span.error(error)
                span.finish()

    # sending heartbeat message to check the readiness
    def send_heartbeat_message(self, topic, msg):
        try:
            if self.kproducer_heartbeat is None:
                _k_endpoint = self.kafka_endpoint

                # Using 2 seconds timeout for heartbeat producer; default of 5 minutes is too long
                self.kproducer_heartbeat = _kafkaProducer(
                    {'bootstrap.servers': _k_endpoint,
                     'default.topic.config' : {'message.timeout.ms': 2000},
                    }
                )

            log.debug('sending-kafka-heartbeat-message', topic=topic)
            # msgs = [msg]

            self.kproducer_heartbeat.produce(topic, msg, callback=self.handle_kafka_delivery_report)

        except Exception as e:
            self.faulty = True
            self.alive_state_handler.callback(self.alive)
            log.error('failed-to-send-kafka-heartbeat-msg', e=e)

    def check_heartbeat_delivery(self):
        try:
            if self.kproducer_heartbeat is not None:
                _msg = self.kproducer_heartbeat.poll(0)
        except Exception as e:
            log.error('failed-to-check-heartbeat-msg-delivery', e=e)
            self.faulty = True

    def handle_kafka_delivery_report(self, err, msg):
        if err is not None:
            # Log and notify only in event of alive status change
            if self.alive is True:
                log.info('failed-to-deliver-message', msg=msg.value(), err=err.str())
                self.alive_state_handler.callback(False)
            self.alive = False
        else:
            if self.alive is not True:
                log.debug('message-delivered-successfully', msg=msg.value())
                self.alive_state_handler.callback(True)
            self.alive = True

    def register_alive_state_update(self, defer_handler):
        self.alive_state_handler = defer_handler

    def is_faulty(self):
        return self.faulty


# Common method to get the singleton instance of the kafka proxy class
def get_kafka_proxy():
    return KafkaProxy._kafka_instance



# Copyright 2017-present Adtran, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from voltha_protos.events_pb2 import EventCategory, EventSubCategory

from pyvoltha_min.adapters.extensions.events.adapter_events import DeviceEventBase


class OnuDiscoveryEvent(DeviceEventBase):
    def __init__(self, event_mgr, pon_id, onu_id, serial_number, onu_device_id, raised_ts):
        super().__init__(event_mgr, raised_ts, object_type='ONU Discovery',
                         event='ONU_DISCOVERY',
                         resource_id=pon_id,
                         category=EventCategory.EQUIPMENT,
                         sub_category=EventSubCategory.PON)
        self._pon_id = pon_id
        self._onu_id = onu_id
        self._serial_number = serial_number
        self._onu_device_id = onu_device_id

    def get_context_data(self):
        return {
            'onu-id': str(self._onu_id),
            'intf-id': str(self._pon_id),
            'serial-number': str(self._serial_number),
            'onu-device-id': str(self._onu_device_id)
        }



#
# Copyright 2016 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
import re

from mock import Mock
from mock import call
from twisted.internet.defer import DeferredQueue, inlineCallbacks
from twisted.trial.unittest import TestCase

from pyvoltha_min.common.event_bus import EventBusClient, EventBus
from six.moves import range


class TestEventBus(TestCase):

    def test_subscribe(self):

        ebc = EventBusClient()
        sub = ebc.subscribe('news', lambda msg, topic: None)
        self.assertEqual(len(ebc.list_subscribers()), 1)
        self.assertEqual(len(ebc.list_subscribers('news')), 1)
        self.assertEqual(len(ebc.list_subscribers('other')), 0)

    def test_unsubscribe(self):

        ebc = EventBusClient(EventBus())
        sub = ebc.subscribe('news', lambda msg, topic: None)
        ebc.unsubscribe(sub)
        self.assertEqual(ebc.list_subscribers(), [])
        self.assertEqual(ebc.list_subscribers('news'), [])

    def test_simple_publish(self):

        ebc = EventBusClient(EventBus())

        mock = Mock()
        ebc.subscribe('news', mock)

        ebc.publish('news', 'message')

        self.assertEqual(mock.call_count, 1)
        mock.assert_called_with('news', 'message')

    def test_topic_filtering(self):

        ebc = EventBusClient(EventBus())

        mock = Mock()
        ebc.subscribe('news', mock)

        ebc.publish('news', 'msg1')
        ebc.publish('alerts', 'msg2')
        ebc.publish('logs', 'msg3')

        self.assertEqual(mock.call_count, 1)
        mock.assert_called_with('news', 'msg1')

    def test_multiple_subscribers(self):

        ebc = EventBusClient(EventBus())

        mock1 = Mock()
        ebc.subscribe('news', mock1)

        mock2 = Mock()
        ebc.subscribe('alerts', mock2)

        mock3 = Mock()
        ebc.subscribe('logs', mock3)

        mock4 = Mock()
        ebc.subscribe('logs', mock4)

        ebc.publish('news', 'msg1')
        ebc.publish('alerts', 'msg2')
        ebc.publish('logs', 'msg3')

        self.assertEqual(mock1.call_count, 1)
        mock1.assert_called_with('news', 'msg1')

        self.assertEqual(mock2.call_count, 1)
        mock2.assert_called_with('alerts', 'msg2')

        self.assertEqual(mock3.call_count, 1)
        mock3.assert_called_with('logs', 'msg3')

        self.assertEqual(mock4.call_count, 1)
        mock4.assert_called_with('logs', 'msg3')

    def test_predicates(self):

        ebc = EventBusClient(EventBus())

        get_foos = Mock()
        ebc.subscribe('', get_foos, lambda msg: msg.startswith('foo'))

        get_bars = Mock()
        ebc.subscribe('', get_bars, lambda msg: msg.endswith('bar'))

        get_all = Mock()
        ebc.subscribe('', get_all)

        get_none = Mock()
        ebc.subscribe('', get_none, lambda msg: msg.find('zoo') >= 0)

        errored = Mock()
        ebc.subscribe('', errored, lambda msg: 1/0)

        ebc.publish('', 'foo')
        ebc.publish('', 'foobar')
        ebc.publish('', 'bar')

        c = call

        self.assertEqual(get_foos.call_count, 2)
        get_foos.assert_has_calls([c('', 'foo'), c('', 'foobar')])

        self.assertEqual(get_bars.call_count, 2)
        get_bars.assert_has_calls([c('', 'foobar'), c('', 'bar')])

        self.assertEqual(get_all.call_count, 3)
        get_all.assert_has_calls([c('', 'foo'), c('', 'foobar'), c('', 'bar')])

        get_none.assert_not_called()

        errored.assert_not_called()

    def test_wildcard_topic(self):

        ebc = EventBusClient(EventBus())
        subs = []

        wildcard_sub = Mock()
        subs.append(ebc.subscribe(re.compile(r'.*'), wildcard_sub))

        prefix_sub = Mock()
        subs.append(ebc.subscribe(re.compile(r'ham.*'), prefix_sub))

        contains_sub = Mock()
        subs.append(ebc.subscribe(re.compile(r'.*burg.*'), contains_sub))

        ebc.publish('news', 1)
        ebc.publish('hamsters', 2)
        ebc.publish('hamburgers', 3)
        ebc.publish('nonsense', 4)

        c = call

        self.assertEqual(wildcard_sub.call_count, 4)
        wildcard_sub.assert_has_calls([
            c('news', 1),
            c('hamsters', 2),
            c('hamburgers', 3),
            c('nonsense', 4)])

        self.assertEqual(prefix_sub.call_count, 2)
        prefix_sub.assert_has_calls([
            c('hamsters', 2),
            c('hamburgers', 3)])

        self.assertEqual(contains_sub.call_count, 1)
        contains_sub.assert_has_calls([c('hamburgers', 3)])

        for sub in subs:
            ebc.unsubscribe(sub)

        self.assertEqual(ebc.list_subscribers(), [])

    @inlineCallbacks
    def test_deferred_queue_receiver(self):

        ebc = EventBus()

        queue = DeferredQueue()

        ebc.subscribe('', lambda _, msg: queue.put(msg))

        for i in range(10):
            ebc.publish('', i)

        self.assertEqual(len(queue.pending), 10)
        for i in range(10):
            msg = yield queue.get()
            self.assertEqual(msg, i)
        self.assertEqual(len(queue.pending), 0)

    def test_subscribers_that_unsubscribe_when_called(self):
        # VOL-943 bug fix check
        ebc = EventBusClient(EventBus())

        class UnsubscribeWhenCalled(object):
            def __init__(self):
                self.subscription = ebc.subscribe('news', self.unsubscribe)
                self.called = False

            def unsubscribe(self, _topic, _msg):
                self.called = True
                ebc.unsubscribe(self.subscription)

        ebc1 = UnsubscribeWhenCalled()
        ebc2 = UnsubscribeWhenCalled()
        ebc3 = UnsubscribeWhenCalled()

        ebc.publish('news', 'msg1')

        self.assertTrue(ebc1.called)
        self.assertTrue(ebc2.called)
        self.assertTrue(ebc3.called)



# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import structlog
from opentracing import global_tracer, Format, Tracer, tags
from jaeger_client import Span as JaegerSpan

ROOT_SPAN_KEY_NAME = "op-name"

log = structlog.get_logger()


class _TracingSupport:
    def __init__(self):
        self._log_correlation_enabled = None
        self._trace_publishing_enabled = None
        self._component_name = None
        self._active_trace_agent_address = None

    @property
    def log_correlation_status(self):
        return self._log_correlation_enabled is not None and self._log_correlation_enabled

    @log_correlation_status.setter
    def log_correlation_status(self, value):
        if not isinstance(value, bool):
            raise TypeError('Boolean required')
        if self._log_correlation_enabled is not None:
            raise ValueError('Log Correlation has already been set to {}'.format(self._log_correlation_enabled))

        self._log_correlation_enabled = value

    @property
    def trace_publishing_status(self):
        return self._trace_publishing_enabled is not None and self._trace_publishing_enabled

    @trace_publishing_status.setter
    def trace_publishing_status(self, value):
        if not isinstance(value, bool):
            raise TypeError('Boolean required')
        if self._trace_publishing_enabled is not None:
            raise ValueError('Trace publishing has already been set to {}'.format(self._trace_publishing_enabled))

        self._trace_publishing_enabled = value

    @property
    def component_name(self):
        return self._component_name is not None and self._component_name

    @component_name.setter
    def component_name(self, value):
        if not isinstance(value, str):
            raise TypeError('String required')
        if self._component_name is not None:
            raise ValueError('Component name has already been set to {}'.format(self._component_name))

        self._component_name = value

    @property
    def active_trace_agent_address(self):
        return self._active_trace_agent_address is not None and self._active_trace_agent_address

    @active_trace_agent_address.setter
    def active_trace_agent_address(self, value):
        if not isinstance(value, str):
            raise TypeError('String required')
        if self._active_trace_agent_address is not None:
            raise ValueError('Active Trace Agent address has already been set to {}'.
                             format(self._active_trace_agent_address))

        self._active_trace_agent_address = value


GlobalTracingSupport = _TracingSupport()


class NilSpan:
    """ Span that is not an opentrace span """
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        NilSpan._on_error(self, exc_type, exc_val, exc_tb)

    def __str__(self):
        return 'NilSpan'

    @staticmethod
    def _on_error(span, exc_type, exc_val, exc_tb):
        if span and exc_val:
            log.info('exception', message=str(exc_val), kind=exc_type, stack=str(exc_tb))


class NilScope:
    """ Scope that is not an opentrace scope """
    def __init__(self):
        self.span = NilSpan()

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        NilSpan._on_error(self.span, exc_type, exc_val, exc_tb)

    def __str__(self):
        return 'NilScope'


def create_async_span(ctx, span_name):
    return None, None

    # // Creates a Async Child Span with Follows-From relationship from Parent Span embedded in passed context.
    # // Should be used only in scenarios when
    # // a) There is dis-continuation in execution and thus result of Child span does not affect the Parent flow at all
    # // b) The execution of Child Span is guaranteed to start after the completion of Parent Span
    # // In case of any confusion, use CreateChildSpan method
    # // Some situations where this method would be suitable includes Kafka Async RPC call, Propagation of Event across
    # // a channel etc.
    # func CreateAsyncSpan(ctx context.Context, taskName string, keyAndValues ...Fields) (opentracing.Span, context.Context) {
    #     if !GetGlobalLFM().GetLogCorrelationStatus() && !GetGlobalLFM().GetTracePublishingStatus() {
    #         return opentracing.NoopTracer{}.StartSpan(taskName), ctx
    #     }
    #
    #     var asyncSpan opentracing.Span
    #     var newCtx context.Context
    #
    #     parentSpan := opentracing.SpanFromContext(ctx)
    #
    #     // We should always be creating Aysnc span from a Valid parent span. If not, create a Child span instead
    #     if parentSpan == nil {
    #         logger.Warn(context.Background(), "Async span must be created with a Valid parent span only")
    #         asyncSpan, newCtx = opentracing.StartSpanFromContext(ctx, taskName)
    #     } else {
    #         // Use Background context as the base for Follows-from case; else new span is getting both Child and FollowsFrom relationship
    #         asyncSpan, newCtx = opentracing.StartSpanFromContext(context.Background(), taskName, opentracing.FollowsFrom(parentSpan.Context()))
    #     }
    #
    #     if parentSpan == nil || parentSpan.BaggageItem(RootSpanNameKey) == "" {
    #         asyncSpan.SetBaggageItem(RootSpanNameKey, taskName)
    #     }
    #
    #     EnrichSpan(newCtx, keyAndValues...)
    #     return asyncSpan, newCtx
    # }


def create_child_span(ctx, task_name, **kwargs):
    """
    Creates a Child Span from Parent Span embedded in passed context. Should be used before starting a new major
    operation in Synchronous or Asynchronous mode (go routine), such as following:

        1. Start of all implemented External API methods unless using a interceptor for auto-injection
           of Span (Server side impl)
        2. Just before calling an Third-Party lib which is invoking a External API (etcd, kafka)
        3. In start of a Go Routine responsible for performing a major task involving significant duration
        4. Any method which is suspected to be time consuming...

    """
    tracer = global_tracer()
    if tracer is None or not (GlobalTracingSupport.log_correlation_status or
                              GlobalTracingSupport.trace_publishing_status):
        # return nop tracer
        return Tracer().start_span(task_name), ctx

    parent_span = tracer.active_span
    child_span = tracer.start_span(task_name, child_of=parent_span)
    #     parentSpan := opentracing.SpanFromContext(ctx)
    #     childSpan, newCtx := opentracing.StartSpanFromContext(ctx, taskName)

    if parent_span is None or not parent_span.get_baggage_item(ROOT_SPAN_KEY_NAME):
        child_span.set_baggage_item(ROOT_SPAN_KEY_NAME, task_name)

    enrich_span(child_span, **kwargs)
    return child_span, ctx
    #     EnrichSpan(newCtx, keyAndValues...)
    #     return childSpan, newCtx
    # }


def enrich_span(span, **kwargs):
    """ Method to inject additional log fields into Span e.g. device-id """

    if isinstance(span, JaegerSpan):
        # Inject as a BaggageItem when the Span is the Root Span so that it propagates
        # across the components along with Root Span (called as Trace)
        # else, inject as a Tag so that it is attached to the Child Task
        is_root_span = span.trace_id == span.span_id

        for _, field in kwargs.items():
            for key, value in field.items():
                if is_root_span:
                    span.set_baggage_item(key, str(value))
                else:
                    span.set_tag(key, value)

# // Extracts details of Execution Context as log fields from the Tracing Span injected into the
# // context instance. Following log fields are extracted:
# // 1. Operation Name : key as 'op-name' and value as Span operation name
# // 2. Operation Id : key as 'op-id' and value as 64 bit Span Id in hex digits string
# //
# // Additionally, any tags present in Span are also extracted to use as log fields e.g. device-id.
# //
# // If no Span is found associated with context, blank slice is returned without any log fields
# func (lfm *LogFeaturesManager) ExtractContextAttributes(ctx context.Context) []interface{} {
# 	if !lfm.isLogCorrelationEnabled {
# 		return make([]interface{}, 0)
# 	}
#
# 	attrMap := make(map[string]interface{})
#
# 	if ctx != nil {
# 		if span := opentracing.SpanFromContext(ctx); span != nil {
# 			if jspan, ok := span.(*jtracing.Span); ok {
# 				// Add Log fields for operation identified by Root Level Span (Trace)
# 				opId := fmt.Sprintf("%016x", jspan.SpanContext().TraceID().Low) // Using Sprintf to avoid removal of leading 0s
# 				opName := jspan.BaggageItem(RootSpanNameKey)
#
# 				taskId := fmt.Sprintf("%016x", uint64(jspan.SpanContext().SpanID())) // Using Sprintf to avoid removal of leading 0s
# 				taskName := jspan.OperationName()
#
# 				if opName == "" {
# 					span.SetBaggageItem(RootSpanNameKey, taskName)
# 					opName = taskName
# 				}
#
# 				attrMap["op-id"] = opId
# 				attrMap["op-name"] = opName
#
# 				// Add Log fields for task identified by Current Span, if it is different
# 				// than operation
# 				if taskId != opId {
# 					attrMap["task-id"] = taskId
# 					attrMap["task-name"] = taskName
# 				}
#
# 				for k, v := range jspan.Tags() {
# 					// Ignore the special tags added by Jaeger, middleware (sampler.type, span.*) present in the span
# 					if strings.HasPrefix(k, "sampler.") || strings.HasPrefix(k, "span.") || k == "component" {
# 						continue
# 					}
#
# 					attrMap[k] = v
# 				}
#
# 				processBaggageItems := func(k, v string) bool {
# 					if k != "rpc-span-name" {
# 						attrMap[k] = v
# 					}
# 					return true
# 				}
#
# 				jspan.SpanContext().ForeachBaggageItem(processBaggageItems)
# 			}
# 		}
# 	}
#
# 	return serializeMap(attrMap)
# }

def mark_span_error(ctx, error):
    """ Inject error information into the span """
    tracing = global_tracer()
    span = tracing.active_span if tracing is not None else None

    if span is not None:
        span.set_tag(tags.ERROR, True)
        span.set_tag('err', error)

# func MarkSpanError(ctx context.Context, err error) {
# 	span := opentracing.SpanFromContext(ctx)
# 	if span != nil {
# 		span.SetTag("error", true)
# 		span.SetTag("err", err)
# 	}
# }



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Resource KV store - interface between Resource Manager and backend store."""
import structlog

from pyvoltha_min.common.config.config_backend import EtcdStore

# KV store uses this prefix to store resource info
PATH_PREFIX = 'service/voltha/resource_manager/{}'


class ResourceKvStore(object):
    """Implements apis to store/get/remove resource in backend store."""

    def __init__(self, technology, device_id, backend, host, port):
        """
        Create ResourceKvStore object.

        Based on backend ('consul' and 'etcd' use the host and port
        to create the respective object.

        :param technology: PON technology
        :param device_id: OLT device id
        :param backend: Type of backend storage (etcd or consul)
        :param host: host ip info for backend storage
        :param port: port for the backend storage
        :raises exception when invalid backend store passed as an argument
        """
        # logger
        self._log = structlog.get_logger()

        path = PATH_PREFIX.format(technology)
        try:
            if backend == 'etcd':
                self._kv_store = EtcdStore(host, port, path)
            else:
                self._log.error('Invalid-backend')
                raise Exception("Invalid-backend-for-kv-store")
        except Exception as e:
            self._log.exception("exception-in-init")
            raise Exception(e)

    def update_to_kv_store(self, path, resource):
        """
        Update resource.

        :param path: path to update the resource
        :param resource: updated resource
        """
        try:
            self._kv_store[path] = str(resource)
            self._log.debug("Resource-updated-in-kv-store", path=path)
            return True
        except BaseException:
            self._log.exception("Resource-update-in-kv-store-failed",
                                path=path, resource=resource)
        return False

    def get_from_kv_store(self, path):
        """
        Get resource.

        :param path: path to get the resource
        """
        resource = None
        try:
            resource = self._kv_store[path]
            self._log.debug("Got-resource-from-kv-store", path=path)
        except KeyError:
            self._log.debug("Resource-not-found-updating-resource",
                            path=path)
        except BaseException:
            self._log.exception("Getting-resource-from-kv-store-failed",
                                path=path)
        return resource

    def remove_from_kv_store(self, path):
        """
        Remove resource.

        :param path: path to remove the resource
        """
        try:
            del self._kv_store[path]
            self._log.debug("Resource-deleted-in-kv-store", path=path)
            return True
        except BaseException:
            self._log.exception("Resource-delete-in-kv-store-failed",
                                path=path)
        return False



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import copy
import structlog

from twisted.internet.defer import inlineCallbacks, returnValue

from voltha_protos.adapter_pb2 import Adapter
from voltha_protos.device_pb2 import DeviceType, Device

DEFAULT_PARTITION_COUNT = 1117
DEFAULT_REPLICATION_FACTOR = 117
DEFAULT_LOAD = 1.1

log = structlog.get_logger()


class EndpointManager:
    def __init__(self, kv_backend,
                 partition_count=DEFAULT_PARTITION_COUNT,
                 replication_factor=DEFAULT_REPLICATION_FACTOR,
                 load=DEFAULT_LOAD):
        self._partition_count = partition_count        # int
        self._replication_factor = replication_factor  # int
        self._load = load                              # float64
        self._backend = kv_backend                     # *db.Backend
        self._services = dict()                        # map[string]*service
        # Device type below is Type as reported by 'voltctl adapter list'
        self._device_type_service_map = dict()         # Device Type (str) -> map[string]string

    @inlineCallbacks
    def get_endpoint(self, device_id, service_type):
        """
        GetEndpoint is called to get the endpoint to communicate with for a specific
        device and service type.  For now this will return the topic name
        :param device_id:  (str)    Device ID
        :param service_type: (str)  Service Type
        :return: (str) Endpoint
        """
        log.debug('entry', device_id=device_id, service=service_type)

        member = yield self._get_owner(device_id, service_type)
        if member is None:
            log.info('owner-not-found', device_id=device_id, service=service_type)
            returnValue('')

        if member.endpoint is None or len(member.endpoint) == 0:
            log.info('endpoint-is-not-set', device_id=device_id, service=service_type)
            returnValue('')

        returnValue(member.endpoint)

    @inlineCallbacks
    def get_all_service_endpoints(self, service_type):
        """ Get a dict() of replica IDs -> endpoints """
        log.debug('entry', service=service_type)

        service, _device_type = yield self._get_service_and_device_type(service_type)
        if service is None:
            log.info('service-not-found', service=service_type)
            returnValue(None)

        returnValue(service.endpoints.copy())

    @inlineCallbacks
    def is_device_owned_by_service(self, device_id, service_type, replica_number):
        log.debug('entry', device_id=device_id, service=service_type, replica_number=replica_number)

        member = yield self._get_owner(device_id, service_type)
        if member is None:
            log.info('owner-not-found', device_id=device_id, service=service_type)
            returnValue(False)

        returnValue(member.replica == replica_number)

    @inlineCallbacks
    def get_replica_assignment(self, device_id, service_type):
        log.debug('entry', device_id=device_id, service=service_type)

        member = yield self._get_owner(device_id, service_type)
        if member is None:
            log.info('owner-not-found', device_id=device_id, service=service_type)
            returnValue(0)

        returnValue(member.replica)

    @inlineCallbacks
    def _get_owner(self, device_id, service_type):
        log.debug('entry', device_id=device_id, service=service_type)

        service, device_type = yield self._get_service_and_device_type(service_type)
        if service is None:
            log.info('service-not-found', device_id=device_id, service=service_type)
            returnValue(None)

        key = self._make_key(device_id, device_type, service_type)
        returnValue(service.get_owner(key))

    @staticmethod
    def _make_key(device_id, device_type, service_type):
        return '{}_{}_{}'.format(service_type, device_type, device_id).encode('utf-8')

    @inlineCallbacks
    def _get_service_and_device_type(self, service_type):
        log.debug('entry', service=service_type)
        service = self._services.get(service_type)

        if service is None or service.total_replicas != service.number_of_owners:
            loaded = yield self._load_services()
            if not loaded:
                returnValue((None, ''))

            # Try again
            service = self._services.get(service_type)
            if service is None or service.total_replicas != service.number_of_owners:
                returnValue((None, ''))

        for device_type, srv_type in self._device_type_service_map.items():
            if srv_type == service_type:
                returnValue((service, device_type))

        returnValue((None, ''))

    @inlineCallbacks
    def get_device_types(self):
        device_types = dict()
        try:
            blobs = yield self._backend.list('device_types')
            if blobs is None:
                log.error('adapters-not-found')
                returnValue(device_types)

        except Exception as e:
            log.exception('device-lookup-failed', e=e)
            returnValue(device_types)

        for blob, kv_metadata in blobs:
            key = kv_metadata.key.decode('utf-8')
            device_type = DeviceType()
            device_type.ParseFromString(blob)
            device_types[key] = device_type

        returnValue(device_types)

    @inlineCallbacks
    def get_adapters(self):
        adapters = dict()
        try:
            blobs = yield self._backend.list('adapters')
            if blobs is None:
                log.error('adapters-not-found')
                returnValue(adapters)

        except Exception as e:
            log.exception('adapter-lookup-failed', e=e)
            returnValue(adapters)

        # Data is marshalled as proto bytes in the data store
        for blob, kv_metadata in blobs:
            key = kv_metadata.key.decode('utf-8')
            adapter = Adapter()
            adapter.ParseFromString(blob)
            adapters[key] = adapter

        returnValue(adapters)

    @inlineCallbacks
    def get_devices(self):
        devices = dict()
        try:
            blobs = yield self._backend.list('devices')
            if blobs is None:
                log.error('devices-not-found')
                returnValue(devices)

        except Exception as e:
            log.exception('devices-lookup-failed', e=e)
            returnValue(devices)

        # Data is marshalled as proto bytes in the data store
        for blob, kv_metadata in blobs:
            key = kv_metadata.key.decode('utf-8')
            device = Device()
            device.ParseFromString(blob)
            devices[key] = device

        returnValue(devices)

    @inlineCallbacks
    def _load_services(self):
        log.debug('entry')

        if self._backend is None:
            log.error('backend-not-set')
            returnValue(False)

        self._services = dict()
        self._device_type_service_map = dict()

        # Load the adapters
        try:
            blobs = yield self._backend.list('adapters')
            if blobs is None:
                log.error('adapters-not-found')
                returnValue(False)

        except Exception as e:
            log.exception('adapter-lookup-failed', e=e)
            returnValue(False)

        # Data is marshalled as proto bytes in the data store
        for blob, kv_metadata in blobs:
            log.debug('key', key=kv_metadata.key.decode('utf-8'))
            adapter = Adapter()
            adapter.ParseFromString(blob)

            # A valid adapter should have the vendorID set
            if adapter.vendor is not None and adapter.vendor != '':
                if adapter.type not in self._services:
                    self._services[adapter.type] = Service(adapter.type,
                                                           adapter.totalReplicas)

                service = self._services[adapter.type]

                if service.total_replicas < adapter.totalReplicas:
                    service.total_replicas = adapter.totalReplicas

                current_replica = adapter.currentReplica
                endpoint = adapter.endpoint

                service.endpoints[current_replica] = endpoint

                key = self._make_key(current_replica, adapter.type, adapter.type)
                service.add_owner(key, Member(adapter.id, adapter.type, adapter.vendor,
                                              endpoint, adapter.version, adapter.currentReplica))

        # Load the device types
        try:
            blobs = yield self._backend.list('device_types')
            if blobs is None:
                log.error('adapters-not-found')
                returnValue(False)

        except Exception as e:
            log.exception('device-lookup-failed', e=e)
            returnValue(False)

        for blob, kv_metadata in blobs:
            log.debug('key', key=kv_metadata.key.decode('utf-8'))
            device_type = DeviceType()
            device_type.ParseFromString(blob)

            if device_type.id not in self._device_type_service_map:
                self._device_type_service_map[device_type.id] = device_type.adapter
        returnValue(True)


class Member:
    """
    Member is an element in the VOLTHA Adapters list
    """
    def __init__(self, ident, srv_type, vendor, endpoint, version, replica):
        """
        Create a new instance in the member list
        :param ident:    (str) Device ID
        :param srv_type: (str) Adapter name
        :param vendor:   (str) Vendor name
        :param endpoint: (str) Listening topic endpoint (Adapter Type + Current Replica)
        :param version:  (str) Build version
        :param replica:  (int) Current Replica number (1..max)
        """
        self._id = ident
        self._service_type = srv_type
        self._vendor = vendor
        self._endpoint = endpoint or ''
        self._version = version
        self._replica = replica
        log.info('member-init', ident=ident, service=srv_type, endpoint=endpoint)

    def __str__(self):
        return self._endpoint

    @property
    def ident(self):
        return self._id

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def replica(self):
        return self._replica

    @property
    def service_type(self):
        return self._service_type


class Service:
    def __init__(self, ident, total_replicas):
        """
        Define a service
        :param ident: (str) ID
        :param total_replicas: (int) Number of replicas
        """
        self._id = ident
        self._total_replicas = total_replicas
        self._endpoints = dict()   # Replica ID (int) -> Endpoint (str)
        self._member = dict()      # Key -> Member

    def get_owner(self, key):
        return self._member.get(key)

    def add_owner(self, key, member):
        self._member[key] = member

    @property
    def endpoints(self):
        return self._endpoints

    @property
    def owner(self):
        return self._member

    @property
    def total_replicas(self):
        return self._total_replicas

    @total_replicas.setter
    def total_replicas(self, value):
        if value < self._total_replicas:
            raise ValueError('Invalid Total Replicas: {}, must be > {}'.format(value, self._total_replicas))
        self._total_replicas = value

    @property
    def number_of_owners(self):
        return len(self._member)



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This facade handles kafka-formatted messages from the Core, extracts the kafka
formatting and forwards the request to the concrete handler.
"""
import structlog
from twisted.internet.defer import inlineCallbacks, returnValue
from zope.interface import implementer

from pyvoltha_min.adapters.interface import IAdapterInterface
from voltha_protos.inter_container_pb2 import IntType, InterAdapterMessage, StrType, Error, ErrorCode
from voltha_protos.device_pb2 import Device, Port, ImageDownload, SimulateAlarmRequest, PmConfigs
from voltha_protos.openflow_13_pb2 import FlowChanges, FlowGroups, Flows, \
    FlowGroupChanges, ofp_packet_out
from voltha_protos.extensions_pb2 import SingleGetValueRequest
from voltha_protos.voltha_pb2 import OmciTestRequest, FlowMetadata, ValueSpecifier
from pyvoltha_min.adapters.kafka.kafka_inter_container_library import get_messaging_proxy, \
    ARG_FROM_TOPIC

log = structlog.get_logger()


class MacAddressError(BaseException):
    def __init__(self, error):
        self.error = error


class IDError(BaseException):
    def __init__(self, error):
        self.error = error


@implementer(IAdapterInterface)
class AdapterRequestFacade:
    """
    Gate-keeper between CORE and device adapters.

    On one side it interacts with Core's internal model and update/dispatch
    mechanisms.

    On the other side, it interacts with the adapters standard interface as
    defined in
    """

    def __init__(self, adapter, core_proxy):
        self.adapter = adapter
        self.core_proxy = core_proxy

    @inlineCallbacks
    def start(self):
        log.debug('starting')

    @inlineCallbacks
    def stop(self):
        log.debug('stopping')

    def start_omci_test(self, device, omcitestrequest, **_kwargs):
        if not device:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        if not omcitestrequest:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="omcitestrequest-invalid")

        d = Device()
        device.Unpack(d)
        omci_test = OmciTestRequest()
        omcitestrequest.Unpack(omci_test)
        result = self.adapter.start_omci_test(d, omci_test.uuid)
        return True, result

    def adopt_device(self, device, **kwargs):
        d = Device()
        if device:
            device.Unpack(d)

            # Update the core reference for that device as it will be used
            # by the adapter to send async messages to the Core.
            if ARG_FROM_TOPIC in kwargs:
                t = StrType()
                kwargs[ARG_FROM_TOPIC].Unpack(t)
                # Update the core reference for that device
                self.core_proxy.update_device_core_reference(d.id, t.val)

            # # Start the creation of a device specific topic to handle all
            # # subsequent requests from the Core. This adapter instance will
            # # handle all requests for that device.
            # reactor.callLater(0, self.createKafkaDeviceTopic, d.id)

            result = self.adapter.adopt_device(d)
            # return True, self.adapter.adopt_device(d)

            return True, result

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def get_ofp_device_info(self, device, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
            return True, self.adapter.get_ofp_device_info(d)

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def reconcile_device(self, device, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
            return True, self.adapter.reconcile_device(d)

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def abandon_device(self, device, **_kwargs):
        return self.adapter.abandon_device(device)

    def disable_device(self, device, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
            return True, self.adapter.disable_device(d)

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def reenable_device(self, device, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
            return True, self.adapter.reenable_device(d)

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def reboot_device(self, device, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
            return True, self.adapter.reboot_device(d)

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def update_pm_config(self, device, pm_configs, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        pm = PmConfigs()
        if pm_configs:
            pm_configs.Unpack(pm)

        return True, self.adapter.update_pm_config(d, pm)

    def download_image(self, device, request, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        img = ImageDownload()
        if request:
            request.Unpack(img)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="port-no-invalid")

        return True, self.adapter.download_image(device, request)

    def get_image_download_status(self, device, request, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        img = ImageDownload()
        if request:
            request.Unpack(img)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="port-no-invalid")

        return True, self.adapter.get_image_download_status(device, request)

    def cancel_image_download(self, device, request, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        img = ImageDownload()
        if request:
            request.Unpack(img)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="port-no-invalid")

        return True, self.adapter.cancel_image_download(device, request)

    def activate_image_update(self, device, request, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        img = ImageDownload()
        if request:
            request.Unpack(img)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="port-no-invalid")

        return True, self.adapter.activate_image_update(device, request)

    def revert_image_update(self, device, request, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        img = ImageDownload()
        if request:
            request.Unpack(img)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="port-no-invalid")

        return True, self.adapter.revert_image_update(device, request)

    def enable_port(self, deviceId, port, **_kwargs):
        if not deviceId:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="device-id")
        if not port:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="port-invalid")
        t = StrType()
        deviceId.Unpack(t)
        p = Port()
        port.Unpack(p)

        return True, self.adapter.enable_port(t.val, p.port_no)

    def disable_port(self, deviceId, port, **_kwargs):
        if not deviceId:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="device-id")
        if not port:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="port-invalid")
        t = StrType()
        deviceId.Unpack(t)
        p = Port()
        port.Unpack(p)

        return True, self.adapter.disable_port(t.val, p.port_no)

    def child_device_lost(self, pDeviceId, pPortNo, onuID, **_kwargs):
        if not pDeviceId:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="device-id")
        if not pPortNo:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="p-port-no")
        if not onuID:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="onu-id")
        t = StrType()
        pDeviceId.Unpack(t)

        port_no = IntType()
        pPortNo.Unpack(port_no)

        oid = IntType()
        onuID.Unpack(oid)

        return True, self.adapter.child_device_lost(t.val, port_no.val, oid.val)

    def self_test(self, device, **_kwargs):
        return self.adapter.self_test_device(device)

    def delete_device(self, device, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
            result = self.adapter.delete_device(d)
            # return (True, self.adapter.delete_device(d))

            # Before we return, delete the device specific topic as we will no
            # longer receive requests from the Core for that device
            kafka_proxy = get_messaging_proxy()
            device_topic = kafka_proxy.get_default_topic() + "/" + d.id
            kafka_proxy.unsubscribe(topic=device_topic)

            return True, result

        return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                            reason="device-invalid")

    def get_device_details(self, device, **_kwargs):
        return self.adapter.get_device_details(device)

    def update_flows_bulk(self, device, flows, groups, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        f = Flows()
        if flows:
            flows.Unpack(f)

        g = FlowGroups()
        if groups:
            groups.Unpack(g)

        return True, self.adapter.update_flows_bulk(d, f, g)

    def update_flows_incrementally(self, device, flow_changes, group_changes, flow_metadata, **_kwargs):
        dev = Device()
        if device:
            device.Unpack(dev)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        flows = FlowChanges()
        if flow_changes:
            flow_changes.Unpack(flows)

        groups = FlowGroupChanges()
        if group_changes:
            group_changes.Unpack(groups)

        metadata = FlowMetadata()
        if flow_metadata:
            flow_metadata.Unpack(metadata)

        return True, self.adapter.update_flows_incrementally(dev, flows, groups, metadata)

    def suppress_event(self, event_filter, **_kwargs):
        return self.adapter.suppress_event(event_filter)

    def unsuppress_event(self, event_filter, **_kwargs):
        return self.adapter.unsuppress_event(event_filter)

    def process_inter_adapter_message(self, msg, **kwargs):
        m = InterAdapterMessage()
        if msg:
            msg.Unpack(m)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="msg-invalid")
        from_topic = StrType()
        topic = kwargs.get('fromTopic')
        if topic:
            topic.Unpack(from_topic)

        log.debug('rx-message', message=m, from_topic=from_topic)
        return True, self.adapter.process_inter_adapter_message(m, from_topic=from_topic.val)

    def receive_packet_out(self, deviceId, outPort, packet, **_kwargs):
        try:
            d_id = StrType()
            if deviceId:
                deviceId.Unpack(d_id)
            else:
                return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                    reason="device-id-invalid")
            op = IntType()
            if outPort:
                outPort.Unpack(op)
            else:
                return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                    reason="outport-invalid")

            p = ofp_packet_out()
            if packet:
                packet.Unpack(p)
            else:
                return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                    reason="packet-invalid")

            return True, self.adapter.receive_packet_out(d_id.val, op.val, p)

        except Exception as e:
            log.exception("error-processing-receive_packet_out", e=e)

    def simulate_alarm(self, device, request, **_kwargs):
        d = Device()
        if device:
            device.Unpack(d)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        req = SimulateAlarmRequest()
        if request:
            request.Unpack(req)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="simulate-alarm-request-invalid")

        return True, self.adapter.simulate_alarm(d, req)

    def get_ext_value(self, deviceId, device, value_flag, **_kwargs):

        if deviceId:
            t = StrType()
            deviceId.Unpack(t)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETER,
                                reason="device-id")
        if device:
            dev = Device()
            device.Unpack(dev)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="device-invalid")
        if value_flag:
            value = ValueSpecifier()
            value_flag.Unpack(value)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="value-flag-invalid")

        return True, self.adapter.get_ext_value(t.val, dev, value)

    @inlineCallbacks
    def single_get_value_request(self, request, **kwargs):
        req = SingleGetValueRequest()
        if request:
            request.Unpack(req)
        else:
            return False, Error(code=ErrorCode.INVALID_PARAMETERS,
                                reason="request-invalid")
        result = yield self.adapter.single_get_value_request(req)
        res = yield result
        return True, res



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
The superclass for all kafka proxy subclasses.
"""

import structlog
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.defer import TimeoutError as TwistedTimeoutErrorDefer
from twisted.internet.error import TimeoutError as TwistedTimeoutError
from twisted.python import failure
from zope.interface import implementer

from pyvoltha_min.common.utils.deferred_utils import DeferredWithTimeout
from pyvoltha_min.common.utils.registry import IComponent

log = structlog.get_logger()


class KafkaMessagingError(BaseException):
    def __init__(self, error):
        super().__init__()
        self.error = error


@implementer(IComponent)
class ContainerProxy:

    def __init__(self, kafka_proxy, remote_topic, my_listening_topic):
        self.kafka_proxy = kafka_proxy
        self.listening_topic = my_listening_topic
        self.remote_topic = remote_topic
        self.default_timeout = 6

    def start(self):
        log.info('started')
        return self

    @staticmethod
    def stop():
        log.info('stopped')

    @classmethod
    def wrap_request(cls, return_cls):
        def real_wrapper(func):
            @inlineCallbacks
            def wrapper(*args, **kw):
                try:
                    results = yield func(*args, **kw)
                    if isinstance(results, tuple):
                        success = results[0]
                        d = results[1]
                    else:
                        success = False
                        d = None

                    if success:
                        log.debug("successful-response", func=func)
                        if return_cls is not None:
                            rclass = return_cls()
                            if d is not None:
                                d.Unpack(rclass)
                            returnValue(rclass)
                        else:
                            log.debug("successful-response-none", func=func)
                            returnValue(None)
                    else:
                        log.info("unsuccessful-request", func=func, args=args, kw=kw)
                        returnValue(d)

                except Exception as e:
                    log.exception("request-wrapper-exception", func=func, e=e)
                    raise

            return wrapper

        return real_wrapper

    @inlineCallbacks
    def invoke(self, rpc, to_topic=None, reply_topic=None, **kwargs):
        @inlineCallbacks
        def _send_request(rpc, m_callback, to_topic, reply_topic, **kwargs):
            try:
                log.debug("sending-request",
                          rpc=rpc,
                          to_topic=to_topic,
                          reply_topic=reply_topic)
                if to_topic is None:
                    to_topic = self.remote_topic
                if reply_topic is None:
                    reply_topic = self.listening_topic
                result = yield self.kafka_proxy.send_request(rpc=rpc,
                                                             to_topic=to_topic,
                                                             reply_topic=reply_topic,
                                                             callback=None,
                                                             **kwargs)
                if not m_callback.called:
                    m_callback.callback(result)
                else:
                    log.debug('timeout-already-occurred', rpc=rpc)

            except Exception as _e:
                log.exception("failure-sending-request", rpc=rpc, kw=kwargs)
                if not m_callback.called:
                    m_callback.errback(failure.Failure())

        # We are going to resend the request on the to_topic if there is a
        # timeout error. This time the timeout will be longer.  If the second
        # request times out then we will send the request to the default
        # core_topic.
        timeouts = [self.default_timeout,
                    self.default_timeout*2,
                    self.default_timeout]
        retry = 0
        max_retry = 2
        for timeout in timeouts:
            d = DeferredWithTimeout(timeout=timeout)
            _send_request(rpc, d, to_topic, reply_topic, **kwargs)
            try:
                res = yield d
                returnValue(res)

            except (TwistedTimeoutError, TwistedTimeoutErrorDefer) as e:
                if retry == max_retry:
                    log.warn('invoke-timeout', e=e, retry=retry, max_retry=max_retry)
                    raise e

                log.info('invoke-timeout', e=e, retry=retry, max_retry=max_retry)
                retry += 1
                if retry == max_retry:
                    to_topic = self.remote_topic

            except Exception as e:
                log.exception('send-request-failed', e=e)
                raise



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Setting up proper logging for Voltha"""

import json
import logging
import logging.config
from collections import OrderedDict

import structlog
from structlog.stdlib import BoundLogger

try:
    from _thread import get_ident as _get_ident
except ImportError:
    from _dummy_thread import get_ident as _get_ident


class StructuredLogRenderer(object):
    def __call__(self, logger, name, event_dict):
        # in order to keep structured log data in event_dict to be forwarded as
        # is, we need to pass it into the logger framework as the first
        # positional argument.
        args = (event_dict,)
        kwargs = {}
        return args, kwargs


class EncoderFix(json.JSONEncoder):
    def default(self, obj):
        return str(obj)


class JsonRenderedOrderedDict(OrderedDict):
    """Our special version of OrderedDict that renders into string as a dict,
       to make the log stream output cleaner.
    """
    def __repr__(self, _repr_running=None):
        # od.__repr__() <==> repr(od)
        call_key = id(self), _get_ident()
        _repr_running = _repr_running or dict()

        if call_key in _repr_running:
            return '...'

        _repr_running[call_key] = 1
        try:
            if not self:
                return ''

            # Convert to JSON but strip off outside '{}'
            msg = '"msg":"{}",'.format(self.pop('event'))
            json_msg = json.dumps(self, indent=None,
                                  cls=EncoderFix,
                                  separators=(', ', ': '),
                                  sort_keys=True)[1:-1]
            return msg + json_msg

        finally:
            del _repr_running[call_key]


class PlainRenderedOrderedDict(OrderedDict):
    """Our special version of OrderedDict that renders into string as a dict,
       to make the log stream output cleaner.
    """
    def __repr__(self, _repr_running=None):
        """od.__repr__() <==> repr(od)"""
        _repr_running = _repr_running or {}
        call_key = id(self), _get_ident()
        if call_key in _repr_running:
            return '...'
        _repr_running[call_key] = 1
        try:
            if not self:
                return '{}'
            return '{%s}' % ", ".join("%s: %s" % (k, v)
                                      for k, v in self.items())
        finally:
            del _repr_running[call_key]


def setup_logging(log_config, instance_id, verbosity_adjust=0):
    """
    Set up logging such that:
    - The primary logging entry method is structlog
      (see http://structlog.readthedocs.io/en/stable/index.html)
    - By default, the logging backend is Python standard lib logger
    """

    def add_exc_info_flag_for_exception(_, name, event_dict):
        if name == 'exception':
            event_dict['exc_info'] = True
        return event_dict

    def add_instance_id(_, __, event_dict):
        event_dict['instance_id'] = instance_id
        return event_dict

    # Configure standard logging
    logging.config.dictConfig(log_config)
    logging.root.level = verbosity_adjust

    processors = [
        add_exc_info_flag_for_exception,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        add_instance_id,
        structlog.processors.UnicodeDecoder(),
        structlog.processors.JSONRenderer()
    ]
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory(),
                        context_class=JsonRenderedOrderedDict,
                        wrapper_class=BoundLogger,
                        processors=processors)

    # Mark first line of log
    log = structlog.get_logger()

    # log.addLevelName(logging.FATAL, 'fatal')
    # log.addLevelName(logging.ERROR, 'error')
    # log.addLevelName(logging.WARN, 'warn')
    # log.addLevelName(logging.INFO, 'info')
    # log.addLevelName(logging.DEBUG, 'debug')

    log.info("first-line", log_level=logging.root.level, verbosity_adjust=verbosity_adjust)
    return log


def string_to_int(loglevel):
    lev = loglevel.upper()
    if lev == "DEBUG":
        return 10
    if lev == "INFO":
        return 20
    if lev == "WARN":
        return 30
    if lev == "ERROR":
        return 40
    if lev == "FATAL":
        return 50
    return 0


def update_logging(instance_id, verbosity_adjust=0):
    """
    Add the vcore id to the structured logger
    :param vcore_id:  The assigned vcore id
    :return: structure logger
    """
    def add_exc_info_flag_for_exception(_, name, event_dict):
        if name == 'exception':
            event_dict['exc_info'] = True
        return event_dict

    def add_instance_id(_, __, event_dict):
        event_dict['instance_id'] = instance_id
        return event_dict

    logging.root.level = verbosity_adjust

    processors = [
        add_exc_info_flag_for_exception,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        add_instance_id,
        StructuredLogRenderer(),
    ]
    structlog.configure(processors=processors)

    # Mark first line of log
    log = structlog.get_logger()
    log.info("updated-logger")
    return log



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Agent to play gateway between adapters.
"""

import codecs
from uuid import uuid4

import six
import structlog
from twisted.internet.defer import inlineCallbacks, returnValue
from voltha_protos.inter_container_pb2 import InterAdapterHeader, InterAdapterMessage

from .container_proxy import ContainerProxy
from .endpoint_manager import EndpointManager
from ..common.kvstore.twisted_etcd_store import TwistedEtcdStore

log = structlog.get_logger()
KV_STORE_DATA_PATH_PREFIX = "service/voltha"


class AdapterProxy(ContainerProxy):     # pylint: disable=too-few-public-methods
    def __init__(self, kafka_proxy, adapter_topic, my_listening_topic, kv_store_address):
        super().__init__(kafka_proxy, adapter_topic, my_listening_topic)
        # KV store's IP Address and PORT
        host, port = kv_store_address.split(':', 1)
        etcd = TwistedEtcdStore(host, port, KV_STORE_DATA_PATH_PREFIX)
        self._endpoint_manager = EndpointManager(etcd)

    @staticmethod
    def _to_string(unicode_str):
        if unicode_str is not None:
            if isinstance(unicode_str, six.string_types):
                return unicode_str
            return codecs.encode(unicode_str, 'ascii')
        return None

    # pylint: disable=too-many-arguments
    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def send_inter_adapter_message(self, msg, msg_type, from_adapter, to_adapter, endpoint,
                                   to_device_id=None, proxy_device_id=None,
                                   message_id=None):
        """
        Sends a message directly to an adapter. This is typically used to send
        proxied messages from one adapter to another.  An initial ACK response
        is sent back to the invoking adapter.  If there is subsequent response
        to be sent back (async) then the adapter receiving this request will
        use this same API to send back the async response.
        :param msg : GRPC message to send
        :param msg_type : InterAdapterMessageType of the message to send
        :param from_adapter: Name of the adapter making the request.
        :param to_adapter: Name of the remote adapter.
        :param endpoint: Topic Endpoint. If not provided, it will be looked up
        :param to_device_id: The ID of the device for to the message is
        intended. if it's None then the message is not intended to a specific
        device.  Its interpretation is adapter specific.
        :param proxy_device_id: The ID of the device which will proxy that
        message. If it's None then there is no specific device to proxy the
        message.  Its interpretation is adapter specific.
        :param message_id: A unique number for this transaction that the
        adapter may use to correlate a request and an async response.
        """
        try:
            # Set to_adapter
            if to_adapter is None:
                to_adapter = self.remote_topic

            # HACK: If endpoint not provided assume a single replica instance
            if endpoint is None or len(endpoint) == 0:
                endpoint = to_adapter + '_1'

            # Build the inter adapter message
            hdr = InterAdapterHeader()
            hdr.type = msg_type
            hdr.from_topic = self._to_string(from_adapter)
            hdr.to_topic = self._to_string(endpoint)
            hdr.to_device_id = self._to_string(to_device_id)
            hdr.proxy_device_id = self._to_string(proxy_device_id)
            hdr.id = self._to_string(message_id) if message_id else uuid4().hex

            hdr.timestamp.GetCurrentTime()

            ia_msg = InterAdapterMessage()
            ia_msg.header.CopyFrom(hdr)
            ia_msg.body.Pack(msg)

            log.debug("sending-inter-adapter-message", type=ia_msg.header.type,
                      from_topic=ia_msg.header.from_topic, to_topic=ia_msg.header.to_topic,
                      to_device_id=ia_msg.header.to_device_id)

            res = yield self.invoke(rpc="process_inter_adapter_message",
                                    to_topic=ia_msg.header.to_topic,
                                    msg=ia_msg)
            returnValue(res)

        except Exception as e:
            log.warn("error-sending-request", e=e)

    @property
    def endpoint_manager(self):
        return self._endpoint_manager



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Agent to play gateway between CORE and an adapter.
"""
import structlog
from google.protobuf.message import Message
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.defer import TimeoutError as TwistedTimeoutErrorDefer
from twisted.internet.error import TimeoutError as TwistedTimeoutError
from voltha_protos.common_pb2 import ID, ConnectStatus, OperStatus
from voltha_protos.device_pb2 import Device, Ports, Devices
from voltha_protos.inter_container_pb2 import StrType, BoolType, IntType, Packet
from voltha_protos.voltha_pb2 import CoreInstance

from .container_proxy import ContainerProxy

log = structlog.get_logger()


def createSubTopic(*args):
    return '_'.join(args)


class CoreProxy(ContainerProxy):

    def __init__(self, kafka_proxy, default_core_topic, default_event_topic, my_listening_topic):
        super(CoreProxy, self).__init__(kafka_proxy, default_core_topic,
                                        my_listening_topic)
        self.core_default_topic = default_core_topic
        self.event_default_topic = default_event_topic
        self.deviceId_to_core_map = dict()

    def update_device_core_reference(self, device_id, core_topic):
        log.debug("update_device_core_reference")
        self.deviceId_to_core_map[device_id] = core_topic

    def delete_device_core_reference(self, device_id, _core_topic):
        log.debug("delete_device_core_reference")
        del self.deviceId_to_core_map[device_id]

    def get_adapter_topic(self, **_kwargs):
        return self.listening_topic

    def get_core_topic(self, device_id):
        if device_id in self.deviceId_to_core_map:
            return self.deviceId_to_core_map[device_id]
        return self.core_default_topic

    @ContainerProxy.wrap_request(CoreInstance)
    @inlineCallbacks
    def register(self, adapter, deviceTypes):
        log.debug("register")

        if adapter.totalReplicas == 0 and adapter.currentReplica != 0:
            raise Exception("totalReplicas can't be 0, since you're here you have at least one")

        if adapter.currentReplica == 0 and adapter.totalReplicas != 0:
            raise Exception("currentReplica can't be 0, it has to start from 1")

        if adapter.currentReplica == 0 and adapter.totalReplicas == 0:
            # if the adapter is not setting these fields they default to 0,
            # in that case it means the adapter is not ready to be scaled
            # and thus it defaults to a single instance
            adapter.currentReplica = 1
            adapter.totalReplicas = 1

        if adapter.currentReplica > adapter.totalReplicas:
            raise Exception("currentReplica (%d) can't be greater than totalReplicas (%d)"
                            % (adapter.currentReplica, adapter.totalReplicas))

        try:
            res = yield self.invoke(rpc="Register",
                                    adapter=adapter,
                                    deviceTypes=deviceTypes)
            log.info("registration-returned", res=res)
            returnValue(res)

        except (TwistedTimeoutError, TwistedTimeoutErrorDefer):
            log.info('register-timeout', adapter=adapter, deviceTypes=deviceTypes)
            raise   # Let caller get timeout so it can decide its own retransmission pattern

        except Exception as e:
            log.exception("registration-exception", e=e)
            raise

    @ContainerProxy.wrap_request(Device)
    @inlineCallbacks
    def get_device(self, device_id):
        log.debug("get-device")
        dev_id = ID()
        dev_id.id = device_id
        # Once we have a device being managed, all communications between the
        # the adapter and the core occurs over a topic associated with that
        # device
        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="GetDevice",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id)
        returnValue(res)

    @ContainerProxy.wrap_request(Device)
    @inlineCallbacks
    def get_child_device(self, parent_device_id, **kwargs):
        log.debug("get-child-device")
        dev_id = ID()
        dev_id.id = parent_device_id
        to_topic = self.get_core_topic(parent_device_id)
        reply_topic = self.get_adapter_topic()
        args = self._to_proto(**kwargs)
        res = yield self.invoke(rpc="GetChildDevice",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                **args)
        returnValue(res)

    @ContainerProxy.wrap_request(Ports)
    @inlineCallbacks
    def get_ports(self, device_id, port_type):
        dev_id = ID()
        dev_id.id = device_id
        p_type = IntType()
        p_type.val = port_type
        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="GetPorts",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                port_type=p_type)
        returnValue(res)

    @ContainerProxy.wrap_request(Devices)
    @inlineCallbacks
    def get_child_devices(self, parent_device_id):
        log.debug("get-child-devices")
        dev_id = ID()
        dev_id.id = parent_device_id
        to_topic = self.get_core_topic(parent_device_id)
        reply_topic = self.get_adapter_topic()
        res = yield self.invoke(rpc="GetChildDevices",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id)
        returnValue(res)

    @ContainerProxy.wrap_request(Device)
    @inlineCallbacks
    def get_child_device_with_proxy_address(self, proxy_address):
        log.debug("get-child-device-with-proxy-address")
        dev_id = ID()
        dev_id.id = proxy_address.device_id
        to_topic = self.get_core_topic(proxy_address.device_id)
        reply_topic = self.get_adapter_topic()
        res = yield self.invoke(rpc="GetChildDeviceWithProxyAddress",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                proxy_address=proxy_address)
        returnValue(res)

    def _to_proto(self, **kwargs):
        encoded = {}
        for k, v in kwargs.items():
            if isinstance(v, Message):
                encoded[k] = v
            elif type(v) == int:
                i_proto = IntType()
                i_proto.val = v
                encoded[k] = i_proto
            elif type(v) == str:
                s_proto = StrType()
                s_proto.val = v
                encoded[k] = s_proto
            elif type(v) == bool:
                b_proto = BoolType()
                b_proto.val = v
                encoded[k] = b_proto
        return encoded

    @ContainerProxy.wrap_request(Device)
    @inlineCallbacks
    def child_device_detected(self,
                              parent_device_id,
                              parent_port_no,
                              child_device_type,
                              channel_id,
                              **kw):
        dev_id = ID()
        dev_id.id = parent_device_id
        ppn = IntType()
        ppn.val = parent_port_no
        cdt = StrType()
        cdt.val = child_device_type
        channel = IntType()
        channel.val = channel_id
        to_topic = self.get_core_topic(parent_device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, parent_device_id)
        # reply_topic = createSubTopic(self.listening_topic, parent_device_id)
        args = self._to_proto(**kw)
        res = yield self.invoke(rpc="ChildDeviceDetected",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                parent_device_id=dev_id,
                                parent_port_no=ppn,
                                child_device_type=cdt,
                                channel_id=channel,
                                **args)
        returnValue(res)

    # @ContainerProxy.wrap_request(Devices)
    # @inlineCallbacks
    # def child_device_lost(self, parent_device_id):
    #     log.debug("child-device-lost")
    #     id = ID()
    #     id.id = parent_device_id
    #     to_topic = self.get_core_topic(parent_device_id)
    #     reply_topic = self.get_adapter_topic()
    #     res = yield self.invoke(rpc="ChildDevicesLost",
    #                             to_topic=to_topic,
    #                             reply_topic=reply_topic,
    #                             device_id=id)
    #     returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def device_update(self, device):
        log.debug("device_update")
        to_topic = self.get_core_topic(device.id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device.id)
        # reply_topic = createSubTopic(self.listening_topic, device.id)
        res = yield self.invoke(rpc="DeviceUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device=device)
        returnValue(res)

    def child_device_removed(self, parent_device_id, child_device_id):
        raise NotImplementedError()

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def device_state_update(self, device_id,
                            oper_status=None,
                            connect_status=None):
        dev_id = ID()
        dev_id.id = device_id
        o_status = IntType()
        if oper_status or oper_status == OperStatus.UNKNOWN:
            o_status.val = oper_status
        else:
            o_status.val = -1
        c_status = IntType()
        if connect_status or connect_status == ConnectStatus.UNKNOWN:
            c_status.val = connect_status
        else:
            c_status.val = -1

        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        #     reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="DeviceStateUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                oper_status=o_status,
                                connect_status=c_status)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def children_state_update(self, device_id,
                              oper_status=None,
                              connect_status=None):
        dev_id = ID()
        dev_id.id = device_id
        o_status = IntType()
        if oper_status or oper_status == OperStatus.UNKNOWN:
            o_status.val = oper_status
        else:
            o_status.val = -1
        c_status = IntType()
        if connect_status or connect_status == ConnectStatus.UNKNOWN:
            c_status.val = connect_status
        else:
            c_status.val = -1

        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="ChildrenStateUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                oper_status=o_status,
                                connect_status=c_status)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def reconcile_child_devices(self, parent_device_id):

        log.warn("todo", reason='This has not been debugged yet')
        dev_id = ID()
        dev_id.id = parent_device_id

        to_topic = self.get_core_topic(parent_device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="ReconcileChildDevices",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def port_state_update(self,
                          device_id,
                          port_type,
                          port_no,
                          oper_status):
        dev_id = ID()
        dev_id.id = device_id
        pt = IntType()
        pt.val = port_type
        pNo = IntType()
        pNo.val = port_no
        o_status = IntType()
        o_status.val = oper_status

        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="PortStateUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                port_type=pt,
                                port_no=pNo,
                                oper_status=o_status)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def child_devices_state_update(self, parent_device_id,
                                   oper_status=None,
                                   connect_status=None):

        dev_id = ID()
        dev_id.id = parent_device_id
        o_status = IntType()
        if oper_status or oper_status == OperStatus.UNKNOWN:
            o_status.val = oper_status
        else:
            o_status.val = -1
        c_status = IntType()
        if connect_status or connect_status == ConnectStatus.UNKNOWN:
            c_status.val = connect_status
        else:
            c_status.val = -1

        to_topic = self.get_core_topic(parent_device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, parent_device_id)
        # reply_topic = createSubTopic(self.listening_topic, parent_device_id)
        res = yield self.invoke(rpc="child_devices_state_update",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                parent_device_id=dev_id,
                                oper_status=o_status,
                                connect_status=c_status)
        returnValue(res)

    def child_devices_removed(self, parent_device_id):
        raise NotImplementedError()

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def device_pm_config_update(self, device_pm_config, init=False):   # TODO: Core does not have the init parameter
        log.debug("device_pm_config_update")
        b = BoolType()
        b.val = init
        to_topic = self.get_core_topic(device_pm_config.id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_pm_config.id)
        # reply_topic = createSubTopic(self.listening_topic, device_pm_config.id)
        res = yield self.invoke(rpc="DevicePMConfigUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_pm_config=device_pm_config,
                                init=b)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def port_created(self, device_id, port):
        log.debug("port_created")
        proto_id = ID()
        proto_id.id = device_id
        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="PortCreated",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=proto_id,
                                port=port)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def ports_state_update(self,                # TODO: Golang rw-core hdoes not have filter
                           device_id,
                           port_type_filter,
                           oper_status):
        log.debug("ports_state_update", device_id=device_id, oper_status=oper_status)
        dev_id = ID()
        dev_id.id = device_id
        t_filter = IntType()
        t_filter.val = port_type_filter
        o_status = IntType()
        o_status.val = oper_status

        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="PortsStateUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                port_type_filter=t_filter,
                                oper_status=o_status)
        log.debug("ports_state_update_response", device_id=device_id, port_type_filter=port_type_filter, oper_status=oper_status, response=res)
        returnValue(res)

    def port_removed(self, device_id, port):
        raise NotImplementedError()

    def ports_enabled(self, device_id):
        raise NotImplementedError()

    def ports_disabled(self, device_id):
        raise NotImplementedError()

    def ports_oper_status_update(self, device_id, oper_status):
        raise NotImplementedError()

    def image_download_update(self, img_dnld):
        raise NotImplementedError()

    def image_download_deleted(self, img_dnld):
        raise NotImplementedError()

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def send_packet_in(self, device_id, port, packet):
        log.debug("send_packet_in", device_id=device_id)
        proto_id = ID()
        proto_id.id = device_id
        p = IntType()
        p.val = port
        pac = Packet()
        pac.payload = packet
        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()
        # to_topic = createSubTopic(self.core_topic, device_id)
        # reply_topic = createSubTopic(self.listening_topic, device_id)
        res = yield self.invoke(rpc="PacketIn",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=proto_id,
                                port=p,
                                packet=pac)
        returnValue(res)

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def device_reason_update(self, device_id, reason):
        dev_id = ID()
        dev_id.id = device_id
        rsn = StrType()
        rsn.val = reason
        to_topic = self.get_core_topic(device_id)
        reply_topic = self.get_adapter_topic()

        res = yield self.invoke(rpc="DeviceReasonUpdate",
                                to_topic=to_topic,
                                reply_topic=reply_topic,
                                device_id=dev_id,
                                device_reason=rsn)
        returnValue(res)

    # ~~~~~~~~~~~~~~~~~~~ Handle event submissions ~~~~~~~~~~~~~~~~~~~~~

    def filter_alarm(self, _device_id, _alarm_event):
        '''
        TODO
        alarm filtering functionality is not implemented
        in Voltha 1.x
        '''
        log.warn('filter_alarm is not implemented')
        #return
        #alarm_filters = self.root_proxy.get('/alarm_filters')

        # rule_values = {
        #     'id': alarm_event.id,
        #     'type': AlarmEventType.AlarmEventType.Name(alarm_event.type),
        #     'category': AlarmEventCategory.AlarmEventCategory.Name(
        #         alarm_event.category),
        #     'severity': AlarmEventSeverity.AlarmEventSeverity.Name(
        #         alarm_event.severity),
        #     'resource_id': alarm_event.resource_id,
        #     'device_id': device_id
        # }
        #
        # for alarm_filter in alarm_filters:
        #     if alarm_filter.rules:
        #         exclude = True
        #         for rule in alarm_filter.rules:
        #             log.debug("compare-alarm-event",
        #                       key=EventFilterRuleKey.EventFilterRuleKey.Name(
        #                           rule.key),
        #                       actual=rule_values[
        #                           EventFilterRuleKey.EventFilterRuleKey.Name(
        #                               rule.key)].lower(),
        #                       expected=rule.value.lower())
        #             exclude = exclude and \
        #                       (rule_values[
        #                           EventFilterRuleKey.EventFilterRuleKey.Name(
        #                               rule.key)].lower() == rule.value.lower())
        #             if not exclude:
        #                 break
        #
        #         if exclude:
        #             log.info("filtered-alarm-event", alarm=alarm_event)
        #             return True
        #
        # return False

    @inlineCallbacks
    def submit_event(self, event_msg):
        try:
            res = yield self.kafka_proxy._send_kafka_message(self.event_default_topic, event_msg)
            returnValue(res)
        except Exception as e:
            log.exception('failed-event-submission', type=type(event_msg), e=e)



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Adapter abstract base class
"""

import structlog
from twisted.internet import reactor, task
from twisted.internet.defer import succeed
from voltha_protos.adapter_pb2 import Adapter, AdapterConfig
from voltha_protos.common_pb2 import AdminState
from voltha_protos.device_pb2 import DeviceType, DeviceTypes
from voltha_protos.health_pb2 import HealthStatus
from zope.interface import implementer

from .interface import IAdapterInterface

log = structlog.get_logger()


# pylint: disable=too-many-arguments, too-many-public-methods, abstract-method

@implementer(IAdapterInterface)
class IAdapter:
    def __init__(self,
                 core_proxy,
                 adapter_proxy,
                 config,
                 device_handler_class,
                 name,
                 vendor,
                 version,
                 device_type, vendor_id,
                 accepts_bulk_flow_update=True,
                 accepts_add_remove_flow_updates=False,
                 endpoint=None,
                 current_replica=1,
                 total_replicas=1,
                 adapter_type=None):
        log.debug(
            'Initializing adapter: {} {} {}'.format(vendor, name, version))
        self.core_proxy = core_proxy
        self.adapter_proxy = adapter_proxy
        self.config = config
        self.name = name
        self.supported_device_types = [
            DeviceType(
                id=device_type,
                vendor_id=vendor_id,
                adapter=name,
                accepts_bulk_flow_update=accepts_bulk_flow_update,
                accepts_add_remove_flow_updates=accepts_add_remove_flow_updates
            )
        ]
        self.descriptor = Adapter(
            id=self.name,
            vendor=vendor,
            version=version,
            config=AdapterConfig(),
            currentReplica=current_replica,
            totalReplicas=total_replicas,
            endpoint=endpoint,
            type=adapter_type or name
        )
        self.devices_handlers = dict()  # device_id -> Olt/OnuHandler()
        self.device_handler_class = device_handler_class

    def start(self):
        log.info('Starting adapter: {}'.format(self.name))

    def stop(self):
        log.info('Stopping adapter: {}'.format(self.name))

    def adapter_descriptor(self):
        return self.descriptor

    def device_types(self):
        return DeviceTypes(items=self.supported_device_types)

    def health(self):       # pylint: disable=no-self-use
        # return HealthStatus(state=HealthStatus.HealthState.HEALTHY)
        return HealthStatus(state=HealthStatus.HEALTHY)

    def change_master_state(self, master):
        raise NotImplementedError()

    def get_ofp_device_info(self, device):
        log.debug('get_ofp_device_info_start', device_id=device.id)
        ofp_device_info = self.devices_handlers[device.id].get_ofp_device_info(
            device)
        log.debug('get_ofp_device_info_ends', device_id=device.id)
        return ofp_device_info

    def adopt_device(self, device):
        log.debug('adopt_device', device_id=device.id)
        self.devices_handlers[device.id] = self.device_handler_class(self,
                                                                     device.id)
        reactor.callLater(0, self.devices_handlers[device.id].activate, device)
        log.debug('adopt_device_done', device_id=device.id)
        return device

    def reconcile_device(self, device):
        raise NotImplementedError()

    def abandon_device(self, device):
        raise NotImplementedError()

    def disable_device(self, device):
        log.info('disable-device', device_id=device.id)
        reactor.callLater(0, self.devices_handlers[device.id].disable)
        log.debug('disable-device-done', device_id=device.id)
        return device

    def reenable_device(self, device):
        log.info('reenable-device', device_id=device.id)
        reactor.callLater(0, self.devices_handlers[device.id].reenable)
        log.info('reenable-device-done', device_id=device.id)
        return device

    def reboot_device(self, device):
        log.info('reboot-device', device_id=device.id)
        reactor.callLater(0, self.devices_handlers[device.id].reboot)
        log.info('reboot-device-done', device_id=device.id)
        return device

    def download_image(self, device, request):
        raise NotImplementedError()

    def get_image_download_status(self, device, request):
        raise NotImplementedError()

    def cancel_image_download(self, device, request):
        raise NotImplementedError()

    def activate_image_update(self, device, request):
        raise NotImplementedError()

    def revert_image_update(self, device, request):
        raise NotImplementedError()

    def enable_port(self, device_id, port_no):
        raise NotImplementedError()

    def disable_port(self, device_id, port_no):
        raise NotImplementedError()

    def child_device_lost(self, device_id, parent_port_no, onu_id):
        raise NotImplementedError()

    def self_test_device(self, device):
        log.info('self-test', device_id=device.id)
        result = reactor.callLater(0, self.devices_handlers[
            device.id].self_test_device)
        log.info('self-test-done', device_id=device.id)
        return result

    def delete_device(self, device):
        log.info('delete-device', device_id=device.id)
        reactor.callLater(0, self.devices_handlers[device.id].delete)
        log.info('delete-device-done', device_id=device.id)
        return device

    def get_device_details(self, device):
        raise NotImplementedError()

    def update_flows_bulk(self, device, flows, groups):
        log.debug('bulk-flow-update', device_id=device.id,
                 flows=flows, groups=groups)
        reactor.callLater(0, self.devices_handlers[device.id].update_flow_table,
                          flows.items)
        return device

    def update_flows_incrementally(self, device, flow_changes, group_changes, flow_metadata):
        log.debug('incremental-flow-update', device_id=device.id,
                  flows=flow_changes, groups=group_changes, metadata=flow_metadata)

        handler = self.devices_handlers[device.id]
        # Remove flows
        if len(flow_changes.to_remove.items) != 0:
            reactor.callLater(0, handler.remove_from_flow_table,
                              flow_changes.to_remove.items)

        # Add flows
        if len(flow_changes.to_add.items) != 0:
            reactor.callLater(0, handler.add_to_flow_table,
                              flow_changes.to_add.items, flow_metadata)
        return device

    def get_ext_value(self, device_id, device, value_flag):
        log.debug("adapter-get-ext-value", device=device, device_id=device_id)

        if device_id not in self.devices_handlers:
            return succeed(None)

        handler = self.devices_handlers[device.id]

        def success(results):
            log.info('get-ext-value-success', results=results)
            return results

        def failure(reason):
            log.warn('get-ext-value-failure', results=reason)
            return reason

        d = task.deferLater(reactor, 0, handler.get_ext_value(device, value_flag))
        return d.addCallbacks(success, failure)

    def update_pm_config(self, device, pm_config):
        log.debug("adapter-update-pm-config", device=device, pm_config=pm_config)
        handler = self.devices_handlers[device.id]
        if handler:
            reactor.callLater(0, handler.update_pm_config, device, pm_config)

    def process_inter_adapter_message(self, msg, from_topic=None):
        raise NotImplementedError()

    def receive_packet_out(self, device_id, egress_port_no, msg):
        raise NotImplementedError()

    def suppress_event(self, filter):
        raise NotImplementedError()

    def unsuppress_event(self, filter):
        raise NotImplementedError()

    def single_get_value_request(self, request):
        raise NotImplementedError()

    def single_set_value_request(self, request):
        raise NotImplementedError()

    def _get_handler(self, device):
        if device.id in self.devices_handlers:
            handler = self.devices_handlers[device.id]
            if handler is not None:
                return handler
            return None
        return None


class OltAdapter(IAdapter):
    """
    OLT Adapter base class
    """
    def __init__(self,
                 core_proxy,
                 adapter_proxy,
                 config,
                 device_handler_class,
                 name,
                 vendor,
                 version,
                 device_type,
                 accepts_bulk_flow_update=True,
                 accepts_add_remove_flow_updates=False,
                 endpoint=None,
                 current_replica=1,
                 total_replicas=1,
                 adapter_type=None):

        super().__init__(core_proxy=core_proxy,
                         adapter_proxy=adapter_proxy,
                         config=config,
                         device_handler_class=device_handler_class,
                         name=name,
                         vendor=vendor,
                         version=version,
                         device_type=device_type,
                         vendor_id=None,
                         accepts_bulk_flow_update=accepts_bulk_flow_update,
                         accepts_add_remove_flow_updates=accepts_add_remove_flow_updates,
                         endpoint=endpoint,
                         current_replica=current_replica,
                         total_replicas=total_replicas,
                         adapter_type=adapter_type)
        self.logical_device_id_to_root_device_id = dict()

    def reconcile_device(self, device):
        try:
            self.devices_handlers[device.id] = self.device_handler_class(self,
                                                                         device.id)
            # Work only required for devices that are in ENABLED state
            if device.admin_state == AdminState.ENABLED:
                reactor.callLater(0,
                                  self.devices_handlers[device.id].reconcile,
                                  device)
            else:
                # Invoke the children reconciliation which would setup the
                # basic children data structures
                self.core_proxy.reconcile_child_devices(device.id)
            return device
        except Exception as e:
            log.exception('Exception', e=e)

    def send_proxied_message(self, proxy_address, msg):
        log.debug('send-proxied-message', proxy_address=proxy_address, msg=msg)
        handler = self.devices_handlers[proxy_address.device_id]
        handler.send_proxied_message(proxy_address, msg)

    def process_inter_adapter_message(self, msg, from_topic=None):
        log.debug('process-inter-adapter-message', msg=msg, from_topic=from_topic)
        # Unpack the header to know which device needs to handle this message
        handler = None
        if msg.header.proxy_device_id:
            # typical request
            handler = self.devices_handlers[msg.header.proxy_device_id]
        elif msg.header.to_device_id and \
                msg.header.to_device_id in self.devices_handlers:
            # typical response
            handler = self.devices_handlers[msg.header.to_device_id]
        if handler:
            reactor.callLater(0, handler.process_inter_adapter_message, msg, from_topic=from_topic)

    def receive_packet_out(self, device_id, egress_port_no, msg):
        try:
            log.debug('receive_packet_out', device_id=device_id,
                      egress_port=egress_port_no, msg=msg)
            handler = self.devices_handlers[device_id]
            if handler:
                reactor.callLater(0, handler.packet_out, egress_port_no, msg.data)
        except Exception as e:
            log.exception('packet-out-failure', e=e)



# Copyright 2017-present Open Networking Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Monkey patched json_format to allow best effort decoding of Any fields.
Use the additional flag (strict_any_handling=False) to trigger the
best-effort behavior. Omit the flag, or just use the original json_format
module fot the strict behavior.
"""

from google.protobuf import json_format


class _PatchedPrinter(json_format._Printer):

    def __init__(self, including_default_value_fields=False,
                 preserving_proto_field_name=False,
                 strict_any_handling=False):
        super(_PatchedPrinter, self).__init__(including_default_value_fields,
                                              preserving_proto_field_name)
        self.strict_any_handling = strict_any_handling

    def _BestEffortAnyMessageToJsonObject(self, msg):
        try:
            res = self._AnyMessageToJsonObject(msg)
        except TypeError:
            res = self._RegularMessageToJsonObject(msg, {})
        return res


def MessageToDict(message,
                  including_default_value_fields=False,
                  preserving_proto_field_name=False,
                  strict_any_handling=False):
    """Converts protobuf message to a JSON dictionary.

    Args:
      message: The protocol buffers message instance to serialize.
      including_default_value_fields: If True, singular primitive fields,
          repeated fields, and map fields will always be serialized.  If
          False, only serialize non-empty fields.  Singular message fields
          and oneof fields are not affected by this option.
      preserving_proto_field_name: If True, use the original proto field
          names as defined in the .proto file. If False, convert the field
          names to lowerCamelCase.
      strict_any_handling: If True, converion will error out (like in the
          original method) if an Any field with value for which the Any type
          is not loaded is encountered. If False, the conversion will leave
          the field un-packed, but otherwise will continue.

    Returns:
      A dict representation of the JSON formatted protocol buffer message.
    """
    printer = _PatchedPrinter(including_default_value_fields,
                              preserving_proto_field_name,
                              strict_any_handling=strict_any_handling)
    # pylint: disable=protected-access
    return printer._MessageToJsonObject(message)


def MessageToJson(message,
                  including_default_value_fields=False,
                  preserving_proto_field_name=False,
                  strict_any_handling=False):
  """Converts protobuf message to JSON format.

  Args:
    message: The protocol buffers message instance to serialize.
    including_default_value_fields: If True, singular primitive fields,
        repeated fields, and map fields will always be serialized.  If
        False, only serialize non-empty fields.  Singular message fields
        and oneof fields are not affected by this option.
    preserving_proto_field_name: If True, use the original proto field
        names as defined in the .proto file. If False, convert the field
        names to lowerCamelCase.
    strict_any_handling: If True, converion will error out (like in the
        original method) if an Any field with value for which the Any type
        is not loaded is encountered. If False, the conversion will leave
        the field un-packed, but otherwise will continue.

  Returns:
    A string containing the JSON formatted protocol buffer message.
  """
  printer = _PatchedPrinter(including_default_value_fields,
                            preserving_proto_field_name,
                            strict_any_handling=strict_any_handling)
  return printer.ToJsonString(message)


json_format._WKTJSONMETHODS['google.protobuf.Any'] = [
    '_BestEffortAnyMessageToJsonObject',
    '_ConvertAnyMessage'
]

json_format._Printer._BestEffortAnyMessageToJsonObject = \
    json_format._Printer._AnyMessageToJsonObject



# Copyright 2017-present Adtran, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from voltha_protos.events_pb2 import EventCategory, EventSubCategory

from pyvoltha_min.adapters.extensions.events.adapter_events import DeviceEventBase


class OltLosEvent(DeviceEventBase):
    event_name = 'OLT_LOSS_OF_SIGNAL'

    def __init__(self, event_mgr, intf_id, raised_ts):
        super().__init__(event_mgr, raised_ts, object_type='OLT LOS',
                         event=self.event_name,
                         category=EventCategory.COMMUNICATION,
                         sub_category=EventSubCategory.OLT)
        # Added port type to indicate if alarm was on NNI or PON
        self._intf_id = intf_id

    def get_context_data(self):
        return {'intf-id:': str(self._intf_id)}

    def to_dict(self):
        """ Serialize to a dictionary """
        data = super().to_dict()
        data['intf_id'] = self._intf_id
        return data

    @staticmethod
    def from_dict(data, event_mgr):
        return OltLosEvent(event_mgr, data['intf_id'], data['raised_ts'])



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# from __future__ import absolute_import
# import arrow
# from unittest import TestCase, main
# from pyvoltha_min.adapters.kafka.core_proxy import CoreProxy
# from pyvoltha_min.adapters.extensions.events.adapter_events import AdapterEvents
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_activation_fail_event import OnuActivationFailEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_active_event import OnuActiveEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_discovery_event import OnuDiscoveryEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_dying_gasp_event import OnuDyingGaspEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_equipment_event import OnuEquipmentEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_high_rx_optical_power_event import OnuHighRxOpticalEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_high_tx_optical_power_event import OnuHighTxOpticalEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_laser_bias_current_event import OnuLaserBiasEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_laser_eol_event import OnuLaserEolEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_lob_event import OnuLobEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_lopc_mic_error_event import OnuLopcMicErrorEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_lopc_miss_event import OnuLopcMissEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_los_event import OnuLosEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_low_rx_optical_power_event import OnuLowRxOpticalEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_low_tx_optical_power_event import OnuLowTxOpticalEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_selftest_failure_event import OnuSelfTestFailureEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_signal_degrade_event import OnuSignalDegradeEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_signal_fail_event import OnuSignalFailEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_startup_event import OnuStartupEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_temp_red_event import OnuTempRedEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_temp_yellow_event import OnuTempYellowEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_voltage_red_event import OnuVoltageRedEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_voltage_yellow_event import OnuVoltageYellowEvent
# from pyvoltha_min.adapters.extensions.events.device_events.onu.onu_window_drift_event import OnuWindowDriftEvent
#
# DEFAULT_ONU_DEVICE_ID = 'default_onu_mock'
# DEFAULT_PON_ID = 0
# DEFAULT_ONU_ID = 0
# DEFAULT_ONU_SN = 'TEST00000001'
# DEFAULT_OLT_SN = 'ABCDXXXXYYYY'
# DEFAULT_ONU_REG = 'ABCD1234'
#
# core_proxy = CoreProxy(
#                kafka_proxy=None,
#                default_core_topic='rwcore',
#                default_event_topic='voltha.events',
#                my_listening_topic='openonu')
#
# event_mgr = AdapterEvents(core_proxy, DEFAULT_ONU_DEVICE_ID, DEFAULT_ONU_DEVICE_ID, DEFAULT_ONU_SN)
#
#
# class TestOnuActivationFailEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuActivationFailEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                             arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuActiveEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuActiveEvent(event_mgr, DEFAULT_ONU_DEVICE_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                     DEFAULT_ONU_REG, DEFAULT_OLT_SN, arrow.utcnow().timestamp, onu_id=DEFAULT_ONU_ID )
#
#     def test_get_context_data(self):
#
#         expected_dict = {
#             'pon-id': DEFAULT_PON_ID,
#             'onu-id': DEFAULT_ONU_ID,
#             'serial-number': DEFAULT_ONU_SN,
#             'olt-serial-number': DEFAULT_OLT_SN,
#             'device-id': DEFAULT_ONU_DEVICE_ID,
#             'registration-id': DEFAULT_ONU_REG
#         }
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuDiscoveryEvent(TestCase):
#
#         def setUp(self):
#             self.event = OnuDiscoveryEvent(event_mgr, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                                arrow.utcnow().timestamp)
#         def test_get_context_data(self):
#             expected_dict = {'pon-id': DEFAULT_PON_ID,
#                              'serial-number': DEFAULT_ONU_SN,
#                              'device-type': 'onu'}
#
#             self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuDyingGaspEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuDyingGaspEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuEquipmentEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuEquipmentEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuHighRxOpticalEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuHighRxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuHighTxOpticalEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuHighTxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLaserBiasEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLaserBiasEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLaserEolEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLaserEolEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLobEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLobEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLopcMicErrorEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLopcMicErrorEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLopcMissEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLopcMissEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLosEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLosEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLowRxOpticalEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLowRxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuLowTxOpticalEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuLowTxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuSelfTestFailureEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuSelfTestFailureEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuSignalDegradeEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuSignalDegradeEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, 20, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN,
#                          'inverse-bit-error-rate': 20}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuSignalFailEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuSignalFailEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, 20, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN,
#                          'inverse-bit-error-rate': 20}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuStartupEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuStartupEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuTempRedEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuTempRedEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuTempYellowEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuTempYellowEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuVoltageRedEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuVoltageRedEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                        arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuVoltageYellowEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuVoltageYellowEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
# class TestOnuWindowDriftEvent(TestCase):
#
#     def setUp(self):
#         self.event = OnuWindowDriftEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, 10, 20, DEFAULT_ONU_SN,
#                                            arrow.utcnow().timestamp)
#
#     def test_get_context_data(self):
#         expected_dict = {'onu-id': DEFAULT_ONU_ID,
#                          'onu-intf-id': DEFAULT_PON_ID,
#                          'onu-serial-number': DEFAULT_ONU_SN,
#                          'drift': 10,
#                          'new-eqd': 20}
#
#         self.assertEqual(self.event.get_context_data(), expected_dict)
#
#
# if __name__ == '__main__':
#     main()
